"""
This effectively  walks through module import statements recursively to find
all modules that a given one depends on.
It furthermore manages the packaging of newly found dependencies when requested
 
ISSUES: For speed, this does not use pathhooks unless imp.find_module fails.
Consequently, if modules can be found in two different sys.path entries, the order
processed by this module may differ from the python import system  
Entirely arbitrary pathhooks are not supported for now - only ZipImporter 
    (specifically importers with a archive attribute)
    
There are some hacks to deal with transmitting archives -- we coerce archives to be stored
to cloud.archives/archive. 
An eventual goal is to clean up the hackish pathhook support code 


Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

from __future__ import with_statement
import os
import sys

import threading
import modulefinder
import imp
import struct
import marshal
import dis

from ..serialization import cloudpickle
import logging

from .. import cloudconfig as cc
cloudLog = logging.getLogger("Cloud.Transport")

LOAD_CONST = chr(dis.opname.index('LOAD_CONST'))
IMPORT_NAME = chr(dis.opname.index('IMPORT_NAME'))
STORE_NAME = chr(dis.opname.index('STORE_NAME'))
STORE_GLOBAL = chr(dis.opname.index('STORE_GLOBAL'))
STORE_OPS = [STORE_NAME, STORE_GLOBAL]
HAVE_ARGUMENT = chr(dis.HAVE_ARGUMENT)

ZIP_IMPORT = -9 #custom imp-like type code             

class DependencyManager(modulefinder.ModuleFinder):
    """
    Based off of module finder.
    
    Features:
    -IgnoreList to ignore base python packages for performance purposes
    -Timestamp tracking
    -'Snapshots' to determine new modules
    -Warnings on using custom C extensions

    Note: This is not thread safe: The user of this is responsible for locking it down

    TODO: Be smart with using import hooks (get_code)    
    """
    
    @staticmethod
    def format_module_list(unformatted_list):
        """Format ignorelist/whitelist/etc."""
        mod_list = set()
        for module in unformatted_list:
            modname = module.strip()
            if modname[0] == '#' or modname[0] == ';':
                continue
            modname = modname.split('.')
            
            if modname[-1] == '*':
                mod_list.add(tuple(modname[:-1]))
            else:
                mod_list.add(tuple(modname))            
        return mod_list                            
    
    @staticmethod
    def format_ignore_list(unformatted_list):
        """Format the ignore list"""  
        
        builtins = ['__builtin__', '_codecs', '_locale', '_sre', 'array', 'binascii', 'cPickle', 
                    'cStringIO', 'cmath', 'datetime', 'errno', 'fcntl', 'grp', 'imp', 'marshal', 'math', 'mmap',
                    'itertools', 'operator', 'parser', 'posix', 'pwd', 'select',
                    'signal', 'strop', 'sys', 'time', 'unicodedata', 'zlib']
              
        ignore_list = DependencyManager.format_module_list(unformatted_list)
        
        #add builtins:
        for builtin in builtins:
            ignore_list.add((builtin, ))
        
        return ignore_list
    
    def __init__(self, debug=0, excludes=[], replace_paths=[], whitelist=None):
        """Note that path is no longer supported -- this uses sys.path at runtime"""
        modulefinder.ModuleFinder.__init__(self, sys.path, debug, None, replace_paths)
        self.ignoreList = self.format_ignore_list(excludes)         
        self.set_whitelist(whitelist)
        self.lastSnapshot = set() #tracking         
        self.transitError = set() #avoid excessive extension warnings        
        
        # analyze main which is not transmitted
        m = sys.modules['__main__']
        if getattr(m,'__file__', None) and cloudpickle.useForcedImports:
            #print 'injected main %s' % m
            self.inject_module(m)
            #if main imports a.b we might not see that b has been loaded
            #The below is a hack to detect this case:
            #For when main is stdin, see adapter._check_forced_mods
            checkModules = self.modules.keys() + self.get_ignored_modules()
                     
            for mod in checkModules:

                self.msgout(2, "inspect", m)
                if '.' in mod:
                    loadedmod = sys.modules.get(mod)  
                                    
                    if loadedmod:
                        if not hasattr(m, '___pyc_forcedImports__'):
                            m.___pyc_forcedImports__ = set()
                        m.___pyc_forcedImports__.add(loadedmod)

    
    def set_whitelist(self, whitelist):        
        if whitelist != None:
            self.whiteList = self.format_module_list(whitelist)
            
            # __main__ must be whitelisted or else dep analysis fails
            self.whiteList.add(('__main__',)) 
            m = sys.modules['__main__']
            if getattr(m,'__file__', None):                
                dirs , mod_realname = os.path.split(m.__file__)
                searchname = mod_realname.split('.',1)[0]  #extract fully qualified name
                self.whiteList.add((searchname,))
                       
        else:
            self.whiteList = None
        #print 'whitelist set to %s' % self.whiteList
    
    def should_ignore(self, modname):
        """Check ignoreList/whiteList to determine if this module should not be processed"""
        #print 'should ig %s = dep %s' % (modname, self.whiteList)
        modname = tuple(modname.split('.'))
        
        if modname in self.ignoreList:        
            return True
        for i in range(1,len(modname)): # check parents
            tst = modname[0:-i]
            if tst in  self.ignoreList:
                return True
            
        # if a whitelist is present, verify that modname is included
        # This is opposite logic of the ignore (black) list
        if self.whiteList != None: 
            if modname in self.whiteList:        
                return False        
            for i in range(1,len(modname)): # check parents
                tst = modname[0:-i]
                if tst in self.whiteList:
                    return False
            
            # __init__ hackery: Check if test module is a parent of the whitelist and transport if so
            # This has security implications; documentation must clarify this!
            for whitelist_tuple in self.whiteList:
                #print 'subset test', modname, whitelist_tuple
                if len(whitelist_tuple) > len(modname):
                    subset = whitelist_tuple[0:len(modname)]
                    if modname == subset:
                        return False
                
            
            return True # ignore all items not in whitelist
             
        # if no whitelist, accept all
        return False         

    def load_package(self, fqname, pathname, archive_name = None):
        """Fix bug with not passing parent into find_module"""
        self.msgin(2, "load_package", fqname, pathname)
        
        newname = modulefinder.replacePackageMap.get(fqname)
        if newname:
            fqname = newname
        
        if archive_name:  #part of an archive
            m = self.add_module(fqname, filename = archive_name, 
                                path = [pathname]+ modulefinder.packagePathMap.get(fqname, []),
                                is_archive = True)
        else:
            # As per comment in modulefinder, simulate runtime __path__ additions.
            m = self.add_module(fqname, filename= pathname + '/__init__.py', 
                                path = [pathname]+ modulefinder.packagePathMap.get(fqname, []))
            
        #Bug fix.  python2.6 modulefinder doesn't pass parent to find_module
        fp, buf, stuff = self.find_module("__init__", m.__path__, parent = m)
        
        self.load_module(fqname, fp, buf, stuff)
        self.msgout(2, "load_package ->", m)
        return m

    def inject_module(self, mod):
        """High level module adding.
        This adds an actual module from sys.modules into the finder
        """
    
        #print 'inject %s' % mod
        mname = mod.__name__        
        if mname in self.modules:
            return        
        if self.should_ignore(mname):
            return
        
        parent = None
        if mname == '__main__': #special case
            searchnames = []
            dirs , mod_realname = os.path.split(mod.__file__)
            searchname = mod_realname.split('.',1)[0]  #extract fully qualified name
            package = getattr(mod,'__package__',None)
            if package:
                searchnames = [package, searchname]
            else:
                searchnames = [searchname]
                        
        else:
            searchnames = mname.rsplit('.',1)      
            
            #load parents recursively first...
        if len(searchnames) > 1:  #this module has a parent - resolve it
            pkg = searchnames[0]                        
            parent = sys.modules[pkg]
            if pkg not in self.modules:
                self.inject_module(parent)
            path = sys.modules[pkg].__path__
        else:
            path = None
            
        searchname = searchnames[-1]

        try:
            #print 'find %s %s %s' % (searchname, path, parent)
            fp, pathname, stuff = self.find_module(searchname,path,parent=parent)
            self.load_module(mname, fp, pathname, stuff)
        except (ImportError, KeyError): # KeyError can occur with packages that hacked up system (e.g. pydap)
            pass

    
    def add_module(self, fqname, filename, path = None, is_archive = False):
        """Save timestamp here"""
        if fqname in self.modules:
            return self.modules[fqname]
        #print 'pre-adding %s' % fqname
        if not filename: #ignore any builtin or extension
            return 
        
        if is_archive:            
            #module's filename is set to the actual archive
            relfilename = os.path.split(filename)[1]
            #cloudpickle needs to know about this to deserialize correctly:             
        else:
            
            #extract relative path of file from filename
            numsplits = fqname.count('.') + 1
            relfilename = ""
            absfilename = filename
            for i in xrange(numsplits):
                absfilename, tmp = os.path.split(absfilename)
                relfilename = tmp + '/' + relfilename
                if '__init__' in tmp:
                    #additional split as this is a package and __init__ is not in fqname
                    absfilename, tmp = os.path.split(absfilename)
                    relfilename = tmp + '/' + relfilename      
            relfilename = relfilename[:-1] #remove terminating /
            
        self.modules[fqname] = m = modulefinder.Module(fqname, filename, path)
        cloudLog.debug('Dependent module %s found (relfile=%s, path=%s, filename=%s)', 
                       fqname, relfilename, path, filename)
        #scivm: Timestamp module for update checks
        #Note: Must use 'reserved' names as modulefinder.import_module will setattr(parant, child_mod)
        # for instance, a submodule named _c__timestamp would blow away _c_timestamp
        m._c__timestamp = long(os.path.getmtime(filename))
        m._c__is_archive = is_archive
        m._c__relfilename = relfilename
        return m
    
    """Manually try to find name on sys.path_hooks
    Some code taken from python3.1 implib"""
    def _path_hooks(self, path):
        """Search path hooks for a finder for 'path'.
        """
        hooks = sys.path_hooks
        for hook in hooks:
            try:
                finder = hook(path)
                sys.path_importer_cache[path] = finder
                return finder
            except ImportError:
                continue
        return None
    
    def manual_find(self, name, path):
        """Load with pathhooks. Return none if fails to load or if default importer must be used
        Otherwise returns loader object, path_loader_handles"""
        finder = None
        for entry in path:
            try:
                finder = sys.path_importer_cache[entry]
            except KeyError:
                finder = self._path_hooks(entry)
            if finder:
                loader = finder.find_module(name)
                if loader:
                    return loader, entry
        return None, None #nothing found!
    
    def hack_find(self, name, parent=None):
        """Contains a variety of methods to handle modules that are not on sys.path
        These modules were directly injected into sys.modules through a python library
        """
        
        """Hack 1: Django - If current directory is /.../abc, abc can be imported! Done by adding
            abc straight into modules.  Path includes /.../abc, but not /.../
            We only need to worry about reference to /.../abc/__init__.py
            
            Hack 2: Celery forcibly adds cwd to sys.path; just inject the module in
        """
        
        if parent == None:
            mod = sys.modules.get(name)
            if mod:
                filename = getattr(mod, '__file__', None)
                if filename and os.path.exists(filename):
                    try:
                        if '__init__' in filename:                        
                            #return open(filename, 'U'), filename, (os.path.splitext(filename)[1]
                            cloudLog.debug('hack find package resolved %s' % filename)
                            return (None, os.path.split(filename)[0], ('', '', imp.PKG_DIRECTORY))
                        else: # general reference (celery, et al)
                            cloudLog.debug('hack find file resolved %s' % filename)
                            modname, ext = os.path.splitext(filename)
                            if ext == '.py':
                                ftype = imp.PY_SOURCE
                                mode = 'U'
                            elif ext == '.pyc':
                                ftype = imp.PY_COMPILED
                                mode = 'rb'
                            else:
                                cloudLog.debug('hack find unknown file extension %s', filename)
                                return None
                            retval = open(filename, mode), filename, (ext, mode, ftype)
                            #print retval
                            return retval
                    except IOError, i:
                        cloudLog.debug('hack find file/pkg failed to resolve %s due to exception %s', filename, i)
                        pass
                    
        
        return None

    
    def find_module(self, name, path, parent=None):
        """find_module, ignoring ones in ignorelist
        """
        if parent is not None:
            # assert path is not None
            fullname = parent.__name__+'.'+name
        else:
            fullname = name
        #print 'test to ignore %s -- %s -- %s' % (fullname, parent, path)
        
        if fullname == 'signal': #builtin hack
            pass        
        if self.should_ignore(fullname):
            self.msgout(3, "find_module -> Ignored", fullname)
            raise ImportError, name

        if path is None:
            if name in sys.builtin_module_names:
                return (None, None, ("", "", imp.C_BUILTIN))

            path = sys.path
        #print 'imp is scanning for %s at %s' % (name, path)
        try:
            #print 'ld %s %s' % ( name, path )
            return imp.find_module(name, path)
        except ImportError:
            #print 'imp error on %s %s' % ( name, path )
            #try path hooks
            loader, ldpath =  self.manual_find(name,path)
            if not loader:
                hack_result = self.hack_find(name, parent)
                if hack_result:
                    return hack_result
                else:
                    raise
            #We now have a PEP 302 loader object. Internally, we must format it
            
            if not hasattr(loader, 'archive') or not hasattr(loader, 'get_code'):
                if fullname not in self.transitError:
                    cloudLog.warn("Cloud cannot transmit python module '%s'.  \
                    It needs to be imported by a %s path hook, but such a path hook does not provide both the \
                    'archive' and 'get_code' property..  Import errors may result; please see PiCloud documentation." % (fullname, str(loader)))                
                    self.transitError.add(fullname)
                raise

            return (None,  ldpath+'/'+name, (loader, name, ZIP_IMPORT))
    
    def get_ignored_modules(self):
        """Return list of modules that are used but were ignored"""
        ignored = []
        for name in self.badmodules:
            if self.should_ignore(name):
                ignored.append(name)
        return ignored        
    
    def any_missing_maybe(self):
        """Return two lists, one with modules that are certainly missing
        and one with modules that *may* be missing. The latter names could
        either be submodules *or* just global names in the package.

        The reason it can't always be determined is that it's impossible to
        tell which names are imported when "from module import *" is done
        with an extension module, short of actually importing it.
        
        PiCloud: Use ignoreList
        """
        missing = []
        maybe = []
        for name in self.badmodules:
            if self.should_ignore(name):
                continue
            i = name.rfind(".")
            if i < 0:
                missing.append(name)
                continue
            subname = name[i+1:]
            pkgname = name[:i]
            pkg = self.modules.get(pkgname)
            if pkg is not None:
                if pkgname in self.badmodules[name]:
                    # The package tried to import this module itself and
                    # failed. It's definitely missing.
                    missing.append(name)
                elif subname in pkg.globalnames:
                    # It's a global in the package: definitely not missing.
                    pass
                elif pkg.starimports:
                    # It could be missing, but the package did an "import *"
                    # from a non-Python module, so we simply can't be sure.
                    maybe.append(name)
                else:
                    # It's not a global in the package, the package didn't
                    # do funny star imports, it's very likely to be missing.
                    # The symbol could be inserted into the package from the
                    # outside, but since that's not good style we simply list
                    # it missing.
                    missing.append(name)
            else:
                missing.append(name)
        missing.sort()
        maybe.sort()
        return missing, maybe


    def load_module(self, fqname, fp, pathname, file_info):
        suffix, mode, type = file_info
        #PiCloud: Warn on C extensions and __import_        
        self.msgin(2, "load_module", fqname, fp and "fp", pathname)
        if type == ZIP_IMPORT:            
            #archive (as suffix) is an PEP 302 importer that implements archive and get_code
            #pathname is used to access the file within the loader
            archive = suffix
            #mode is the actual name we want to read
            name = mode
            if archive.is_package(name):  #use load_package with archive set
                m = self.load_package(fqname, pathname, archive_name = archive.archive)
                return m
            else:
                try:
                    co = archive.get_code(name)
                except ImportError:
                    cloudLog.warn("Cloud cannot read '%s' within '%s'.  Import errors may result; \
                    please see PiCloud documentation." % (fqname, archive.archive))
                    raise
                m = self.add_module(fqname, archive.archive, is_archive = True)
        else:            
            if type == imp.PKG_DIRECTORY:
                m = self.load_package(fqname, pathname)
                self.msgout(2, "load_module ->", m)
                return m
            elif type == imp.PY_SOURCE:
                try:
                    co = compile(fp.read()+'\n', pathname, 'exec')
                except SyntaxError: #compilation fail.
                    cloudLog.warn("Syntax error in %s.  Import errors may occur in rare situations." % pathname)
                    raise ImportError ("Syntax error in %s" %pathname)
                    
            elif type == imp.PY_COMPILED:
                if fp.read(4) != imp.get_magic():
                    cloudLog.warn("Magic number on %s is invalid.  Import errors may occur in rare situations." % pathname)
                    self.msgout(2, "raise ImportError: Bad magic number", pathname)
                    raise ImportError, "Bad magic number in %s" % pathname
                fp.read(4)
                co = marshal.load(fp)
            elif type == imp.C_EXTENSION:
                if fqname not in self.transitError:
                    cloudLog.warn("Cloud cannot transmit python extension '%s' located at '%s'.  Import errors may result; please see PiCloud documentation." % (fqname, pathname))                
                    self.transitError.add(fqname)
                raise ImportError(fqname)
            else:
                co = None
            m = self.add_module(fqname, filename=pathname)
        if co:
            if self.replace_paths:
                co = self.replace_paths_in_code(co)
            m.__code__ = co
            names = co.co_names        
            if names and '__import__' in names:
                #PiCloud: Warn on __import__
                    cloudLog.warn('__import__ found within %s. Cloud cannot follow these \
dependencies. You MAY see importerror cloud exceptions. For more information, consult the PiCloud manual' 
                    % fqname)            
            self.scan_code(co, m)
        self.msgout(2, "load_module ->", m)
        return m

    def get_updated_snapshot(self):
        """Return any new myMods values since this was last called"""

        outList = []
        new_snapshot = set()
        for modname, modobj in self.modules.items():
            if modname not in self.lastSnapshot and modname not in new_snapshot:
                if modobj._c__is_archive:  #store archive itself as it can be referenced by multiple modname
                    archive = modobj.__file__
                    if archive in self.lastSnapshot or archive in new_snapshot:
                        continue
                    else:
                        new_snapshot.add(archive)
                outList.append((modobj._c__relfilename, modobj._c__timestamp, modobj._c__is_archive))                                    
                new_snapshot.add(modname)
        return outList, new_snapshot
    
    def commit_snapshot(self, new_snapshot):
        """Commit snapshot after successful module transfer"""
        self.lastSnapshot = self.lastSnapshot.union(new_snapshot)
    

class FilePackager(object):
    """This class is responsible for the packaging of files"""
    """This is not thread safe"""
    
    fileCollection = None
    depManager = None
    ARCHIVE_PATH = 'cloud.archive/' #location where archives are extracted
       
    def __init__(self, path_infos=None, dep_manager = None):
        """path_infos is a list of (paths relative to site-packages, archive)""" 
        self.fileCollection = {}
        self.depManager = dep_manager
        if path_infos:
            for relPath, archive in path_infos:
                if archive:
                    self.add_archive(relPath)
                else:
                    self.add_relative_path(relPath)
    
    def add_archive(self, archive_name):
        for site in sys.path:
            if site.endswith(archive_name):
                self.fileCollection[self.ARCHIVE_PATH + archive_name] = site
    
    def add_relative_path(self, relPath):
        """Add a file by relative path to the File Transfer"""
        for site in sys.path:
            if site != '':
                site += '/'
            tst = os.path.join(site,relPath.encode())
            if os.path.exists(tst):
                self.fileCollection[relPath] = tst     
                return           
        from ..cloud import CloudException
        msg = 'FilePackager: %s not found on sys.path, resolving through modules list. sys.path=%s.' % (relPath, sys.path)
        cloudLog.debug(msg)
        if self.depManager:
            for modobj in self.depManager.modules.values():
                if modobj._c__relfilename == relPath:
                    self.fileCollection[relPath] = modobj.__file__
                    return
        msg = 'FilePackager: %s not found on sys.path %s. or sys.modules' % (relPath)
        cloudLog.error(msg)
        raise CloudException(msg)
            
    
    def get_tarball(self):
        try:
            from cStringIO import StringIO
        except ImportError:
            from StringIO import StringIO
        import tarfile
        
        outfile = StringIO()
        tfile = tarfile.open(name='transfer.tar',fileobj=outfile,mode='w')
        tfile.dereference=True
                
        for arcname, fname in self.fileCollection.items():
            tfile.add(name=fname,arcname=arcname,recursive=False)
        tfile.close()
        
        return outfile.getvalue()
            
        



from __future__ import with_statement 
"""
This module is responsible for managing and writing serialization reports

Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

import errno, os, datetime, stat, time, threading    
import shutil

import distutils
import distutils.dir_util

from . import pickledebug
from .serializationhandlers import DebugSerializer
from .. import cloudconfig as cc
from ..cloudlog import cloudLog, purgeDays
from ..util import fix_sudo_path
from pickledebug import DebugPicklingError 

class SerializationReport():
    c = """Path to save object serialization meta-data.
    This path is relative to ~/.scivm/"""    
    serializeLoggingPath = \
        cc.logging_configurable('serialize_logging_path',
                                          default='datalogs/',
                                          comment=c)
    
    
    #k = __import__('f')
    #p = __builtins__.__import__('g')
    
    pid = None #process identifier
    cntLock = None
        
    def __init__(self, subdir = ""):
        """
        Create logging directory with proper path if subdir is set
        """
        if subdir:     
            logpath = os.path.expanduser(os.path.join(cc.baselocation,self.serializeLoggingPath,subdir))
            
            self.purge_old_logs(logpath)
            
            #uses pidgin's log path format        
            date = str(datetime.datetime.today().date())
            date = date.replace(':','-')
            
            time = str(datetime.datetime.today().time())[:8]
            time = time.replace(':','')
            
            timestamp = date + '.' + time
            
            logpath = os.path.join(logpath,timestamp)
                        
            
            try_limit = 10000
            ctr = 0
            basepath = logpath
            
            
            
            while True:
                try:
                    if not distutils.dir_util.mkpath(logpath):
                        raise distutils.errors.DistutilsFileError('retry')
                except distutils.errors.DistutilsFileError, e:
                    if ctr >= try_limit:
                        raise IOError("can't make file %s. Error is %s" % (logpath,str(e)))
                    ctr+=1                    
                    logpath = basepath + '-%d' % ctr  
                else:
                    break
            
            cloudLog.info("Serialization reports will be written to %s " % logpath)                        
            fix_sudo_path(logpath)                
            self.logPath = logpath
                    
        self.pickleCount = {}
        self.cntLock = threading.Lock()
        
    def purge_old_logs(self, logpath):
        """Remove subdirectories with modified time older than purgeDays days"""
        try:
            subdirs = os.listdir(logpath)
        except OSError, e:
            if e.errno != errno.ENOENT:
                cloudLog.debug('Could not purge %s due to %s', logpath, str(e))
            return
        now = time.time()
        allowed_difference = purgeDays * 24 * 3600 #purge days in seconds
        for s in subdirs: #walk through log subdirectories            
            new_dir = os.path.join(logpath,s)
            try:
                stat_result = os.stat(new_dir)
            except OSError:
                cloudLog.warn('Could not stat %s', new_dir, exc_info = True)
                continue
            if stat.S_ISDIR(stat_result.st_mode) and (now - stat_result.st_mtime) > allowed_difference:
                cloudLog.debug('Deleting %s (%s days old)', new_dir, (now - stat_result.st_ctime)/(24*3600))
                try:
                    shutil.rmtree(new_dir)
                except OSError:
                    cloudLog.warn('Could not delete %s', new_dir, exc_info = True)
                     
    
    def update_counter(self, baselogname):
        baselogname = baselogname.replace('<','').replace('>','')
        with self.cntLock:
            cnt = self.pickleCount.get(baselogname,0)
            cnt+=1
            self.pickleCount[baselogname] = cnt
        return cnt
        
    def get_report_file(self, logname, ext, cnt = None, pid = None):
        """Returns the name of a report file with cnt and pid filled in"""
        logname = logname.replace('<','').replace('>','')
       
        mid = ''
        if pid:
            mid += 'P%d.' % pid
        if cnt:
            mid += '%d.' % cnt

        logname = logname % mid
        
        logname+= ext
        
        return os.path.join(self.logPath,logname)
        
    
    def open_report_file(self, logname, ext, cnt = None, pid = None):
        """Open an arbitrary report file with cnt and pid filled in"""
        return file(self.get_report_file(logname, ext, cnt, pid),'w')  
        
    
    """Reporting"""
    def save_report(self, dbgserializer, logname, cnt = None, pid = ''):
        
        if not hasattr(dbgserializer,'write_debug_report'):
            #due to serialization level being cloud.call argument, we might not have
            # a write_debug_report in active serializer, even though this object exists
            return
        
        #HACK for default detection
        if type(pid) == str:
            pid = self.pid
         
        reportf = self.open_report_file(logname, '.xml', cnt, pid) 

        dbgserializer.write_debug_report(reportf)
        reportf.close()
        
        return reportf.name

    

        



"""
Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2012 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

"""
This code is based on a  simple XML-generator# Originally Lars Marius Garshol, September 1998
http://mail.python.org/pipermail/xml-sig/1998-September/000347.html
Changes by Uche Ogbuji April 2003
 *  unicode support: accept encoding argument and use Python codecs
    for correct character output
 *  switch from deprecated string module to string methods
 *  use PEP 8 style
 *  Further modifications by Science VM

Please see Ogbuji's article for background: http://www.xml.com/pub/a/2003/04/09/py-xml.html
Ogbuji has given permission to modify and release this code under LGPL
"""
 

import sys
import codecs

class XmlWriter:

    def __init__(self, out=sys.stdout, encoding="utf-8", indent=u"  ", header=True):
        """
        out      - a stream for the output
        encoding - an encoding used to wrap the output for unicode
        indent   - white space used for indentation
        """
        wrapper = codecs.lookup(encoding)[3]
        self.out = wrapper(out)
        self.stack = []
        self.indent = indent
        if header:
            self.out.write(u'<?xml version="1.0" encoding="%s"?>\n' \
                       % encoding)
        self.needClose = False

    def close(self):
        if self.out is not sys.stdout:
            self.out.close()

    def doctype(self, root, pubid, sysid):
        """
        Create a document type declaration (no internal subset)
        """
        if pubid == None:
            self.out.write(
                u"<!DOCTYPE %s SYSTEM '%s'>\n" % (root, sysid))
        else:
            self.out.write(
                u"<!DOCTYPE %s PUBLIC '%s' '%s'>\n" \
                % (root, pubid, sysid))
    
    def _closeIfNeeded(self):
        if self.needClose:
            self.out.write(u">")
        self.needClose = False
        
    def comment(self, cmt, attrs={}):
        """
        Create a comment with attributes
        """
        self._closeIfNeeded()
        if self.stack:
            self.out.write('\n')
        self.out.write('<!-- ')
        self.out.write(cmt)
        for (a, v) in attrs.items():
            try:
                self.out.write(u" %s='%s'" % (a, self.__escape_attr(v)))
            except UnicodeDecodeError:
                continue        
        self.out.write(' -->')
        if not self.stack:
            self.out.write('\n')

        
    def push(self, elem, attrs={}):        
        """
        Create an element which will have child elements
        """
        self._closeIfNeeded()
        if self.stack:
            self.out.write('\n')
        self.__indent()
        try:
            self.out.write("<" + elem)
        except UnicodeDecodeError:
            elem = 'UNPRINTABLE_OBJECT'
            self.out.write("<" + elem)  
        for (a, v) in attrs.items():
            try:
                self.out.write(u" %s='%s'" % (a, self.__escape_attr(v)))
            except UnicodeDecodeError:
                continue
        if self.stack:
            self.needClose = True
        else:
            self.out.write(u">")
        self.stack.append(elem)

    def elem(self, elem, content, attrs={}):
        """
        Create an element with text content only
        """
        self._closeIfNeeded()
        if self.stack:
            self.out.write('\n')
        self.__indent()
        self.out.write(u"<" + elem)
        for (a, v) in attrs.items():
            self.out.write(u" %s='%s'" % (a, self.__escape_attr(v)))
        self.out.write(u">%s</%s>" \
                       % (self.__escape_cont(content), elem))
        if not self.stack:
            self.out.write('\n')

    def empty(self, elem, attrs={}):
        """
        Create an empty element
        """
        self._closeIfNeeded()
        if self.stack:
            self.out.write('\n')
        self.__indent()
        self.out.write(u"<"+elem)
        for a in attrs.items():
            self.out.write(u" %s='%s'" % a)
        self.out.write(u"/>")
        if not self.stack:
            self.out.write('\n')
        
    def pop(self):
        """
        Close an element started with the push() method
        """
        elem=self.stack[-1]
        del self.stack[-1]

        if self.needClose: #means we never pushed a subelement onto stack:
            self.out.write(u"/>")
            self.needClose = False                        
        else:
            self.out.write('\n')        
            self.__indent()        
            self.out.write(u"</%s>" % elem)
        #if outer level:
        if not self.stack:
            self.out.write('\n')
        
    def flush(self):
        self.out.flush()
    
    def __indent(self):
        self.out.write(self.indent * (len(self.stack) * 2))
    
    def __escape_cont(self, text):
        return text.replace(u"&", u"&amp;")\
               .replace(u"<", u"&lt;")

    def __escape_attr(self, text):
        return text.replace(u"&", u"&amp;") \
               .replace(u"'", u"&apos;").replace(u"<", u"&lt;")
               
"""
Stack Writer
Will only write a single downward path in the xml tree
"""
class XmlStackWriter:

    def __init__(self, xmlwriter):
        """
        xmlwriter      -  real xmlwriter
        """
        self.xmlwriter = xmlwriter
        self.writecmd = []
        self.extraEl = False
    
    def addElement(self, node):        
        if self.extraEl:
            self.writecmd[-1] = node
        else:
            self.writecmd.append(node)
            self.extraEl = True       
        
    def comment(self, cmt, attrs={}):
        newcmd = (self.xmlwriter.comment,(cmt,attrs))
        self.addElement(newcmd)

        
    def push(self, elem, attrs={}):        
        newcmd = (self.xmlwriter.push,(elem,attrs))
        self.addElement(newcmd)
        self.extraEl = False


    def elem(self, elem, content, attrs={}):
        newcmd = (self.xmlwriter.elem,(elem,content,attrs))
        self.addElement(newcmd)
                    
    def empty(self, elem, attrs={}):
        newcmd = (self.xmlwriter.empty,(elem,attrs))
        self.addElement(newcmd)        
        
    def pop(self):
        if self.extraEl:
            self.writecmd.pop()        
        self.writecmd.pop()        
        self.extraEl = False
        
    def flush(self):
        """
        Flush to xml file
        """        
        for cmd, args in self.writecmd:                     
            cmd(*args)
        self.xmlwriter._closeIfNeeded() #prettification hack

"""
Reading:
"""

import xml.parsers.expat

"""
when you get to a tag:
Cases:
1. If tag has the same name as a class (class.xmltagname()) or if it has a !type attribute! that has the same name as a class (class.xmltagname(): instantiate the class
    - If parent is None, assign class to global list
    - If parent is a class, assign to its member variable with the correct name specified by the !name attribute! (default to tag name)
    - If the parent is a list, create a new object and append it
    - If the parent is a dictionary, create a new key/value pair in the dictionary
2. If tag does not have the same name as a class:
    - If parent is None?
    - If the parent is a class, it must have the same tag name or !name attribute! as a member of the parent class (if there is no parent, throw an exception)
    - If the parent is a dictionary, and the tag has a value, add it as a key/value to the dictionary
    - If the parent is a dictionary, and the tag does not have a value, then it is in limbo mode
    - If the parent is a list, and the tag has a value... (not currently specified)
    - If the parent is not yet specified, its a list, unless there is a value attribute

if tag does not have the same name as a class, it can be either a list or a dict


"""

class XmlDataParser:

    path = []
    objpath = []

    root = None

    parser = None
    tagclasses = {}
    
    lasttagname = None

    def __init__(self):
        self.parser = xml.parsers.expat.ParserCreate()

        self.parser.StartElementHandler = self.start_element
        self.parser.EndElementHandler = self.end_element
        self.parser.CharacterDataHandler = self.char_data

    def addTagClass(self, cls):
        self.tagclasses[cls.xmltagname()] = cls

    def parse(self, filename):
        file = open(filename)
        lines = file.readlines()
        xmltext = ''
        for line in lines:
            xmltext += line.strip()        
        self.parser.Parse(xmltext)
        
        #print "root is " + str(self.root)
        return self.root

    # 3 handler functions
    def start_element(self, name, attrs):

        if len(self.objpath) > 0:
            parent = self.objpath[len(self.objpath)-1]
        else:
            parent = None

        if 'type' in attrs:
            tagtype = attrs['type']
        else:
            tagtype = name

        if 'name' in attrs:
            tagname = attrs['name']
        else:
            tagname = name
        
        self.lasttagname = tagname

        #print 'Entering', name
        #print 'OBJPATH', str(self.objpath)
        #print 'PATH', str(self.path)
        #print 'ROOT:', str(self.root)
        #print 'INFO:', tagtype, tagname
        #print 'PARENT:', parent

        self.path.append(name)

        #if tag has the same name as a class
        if tagtype in self.tagclasses:
            #print 'tagtype in tagclass'
            #newobj = self.tagclasses[tagtype]()
            targetclass = self.tagclasses[tagtype]
            newobj = targetclass()

            if parent is None:
                #print '1'
                self.root = newobj
            elif type(parent) is list:
                #print '2'
                parent.append(newobj)
            elif type(parent) is dict:
                #print '3'
                pass
            else:
                #print '4'
                if hasattr(parent, tagname):
                    setattr(parent, tagname, newobj)
                else:
                    raise Exception('Parent object does not have attribute ' + str(tagname) )
            self.objpath.append(newobj)

        else:
            #print 'tagtype not in tagclass'
            if parent is None:
                #print '1'
                self.root = []
                self.objpath.append(self.root)
            elif type(parent) is dict and 'value' in attrs:
                #print '2'
                parent[tagname] = self.convertValueToBestType(attrs['value'])
                self.objpath.append('N/A')
            elif type(parent) is dict and 'value' not in attrs:
                raise Exception('Parent is dict, but child has no value attribute')
            elif type(parent) is list:                                
                raise Exception('Not really supported... ' + str(parent) + " with " + str(tagname))
            else:
                #print '3'

                if hasattr(parent, tagname) and 'value' in attrs:
                    self.objpath.append('N/A')
                    setattr(parent, tagname, self.convertValueToBestType(attrs['value']))
                elif hasattr(parent, tagname):
                    self.objpath.append(getattr(parent, tagname))
                else:
                    raise Exception('Parent object does not have attribute ' + str(tagname))

        #print ''

    # assume value is passed is as a string
    # as it would be when being parsed in an xml file
    def convertValueToBestType(self, value):
        #print value, type(value)
        if not isinstance(value, basestring):
            raise Exception('convertValueToCorrectType given a non-string input')

        try:
            floatval = float(value)
            if floatval % 1.0 == 0.0:
                return int(value)
            else:
                return floatval
        except ValueError, e:
            lowerval = value.lower()
            if lowerval == "true":
                return True
            elif lowerval == "false":
                return False

            return value

    def end_element(self, name):
        #print "Leaving", name
        self.path.pop()
        self.objpath.pop()
        #if len(self.objpath) > 0:
        #    parent = self.objpath[len(self.objpath)-1]
        #    #print parent
        #print '' 

    def char_data(self, data):
        if len(self.objpath) > 1:
            parent = self.objpath[len(self.objpath)-2]
        else:
            parent = None
        #print 'Character data:', repr(data)
        if hasattr(parent, self.lasttagname):            
            setattr(parent, self.lasttagname, self.convertValueToBestType(str(data)))
        else:        
            raise Exception("cannot handle data < " + repr(data) + "> inside " + str(self.lasttagname))


#p.Parse("""<?xml version="1.0"?>
#<parent id="top"><bob id="asfaf" /><child1 name="paul">Text goes here</child1>
#<child2 name="fred">More text</child2>
#</parent>""", 1)





"""
For managing crons on Science VM.  
Crons allow you to "register" a function to be invoked periodically on PiCloud
according to a schedule you specify.

Api keys must be configured before using any functions in this module.
(via cloudconf.py, cloud.setkey, or being on Science VM server)
"""
from __future__ import absolute_import
"""
Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2011 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

from . import _getcloud, _getcloudnetconnection

from .util import min_args
from .util.zip_packer import Packer
from .util.cronexpr import CronTime

_register_query = 'cron/register/'
_deregister_query = 'cron/deregister/'
_run_query = 'cron/run/'
_list_query = 'cron/list/'
_info_query = 'cron/info/'

"""
This module utilizes the cloud object extensively
The functions can be viewed as instance methods of the Cloud (hence accessing of protected variables)
"""

def _low_level_register(func, label, schedule, params):
    """low level register
    Same as regular register but accepts params rather than kwargs
    """
    if not callable(func):
        raise TypeError( 'cloud.cron.register first argument (%s) is not callable'  % (str(func) ))
    try:
        nargs = min_args(func)
    except TypeError: #some types we cannot error check
        nargs = 0         
    if nargs != 0:
        raise ValueError('cron functions must have 0 (required) arguments. %s requires %s' \
                         % (str(func), nargs))
    
    #check for cron schdule sanity
    if hasattr(schedule,'__iter__'):
        schedule = ' '.join(schedule)
    if not isinstance(schedule, str):
        schedule = str(schedule)
    CronTime(schedule) #throw exception if invalid cron format
    
    #ASCII label:
    try:
        label = label.decode('ascii').encode('ascii')
    except (UnicodeDecodeError, UnicodeEncodeError):
        raise TypeError('label must be an ASCII string')
    
    cloud = _getcloud()
    conn = _getcloudnetconnection()
        
    os_env_vars = params.pop('os_env_vars', None)
    sfunc, sarg, logprefix, logcnt = cloud.adapter.cloud_serialize(func, 2, [],logprefix='cron.',
                                                                   os_env_vars=os_env_vars)
    
    #Below is derived from HttpAdapter.job_add
    conn._update_params(params)
    
    cloud.adapter.dep_snapshot() #let adapter make any needed calls for dep tracking
    
    data = Packer()
    data.add(sfunc)
    params['data'] = data.finish()
    
    params['label'] = label
    params['cron_exp'] = schedule                     
    
    conn.send_request(_register_query, params) 
    

def register(func, label, schedule, **kwargs):
    """       
    Register *func* (a callable) to be run periodically on PiCloud according to *schedule*
    
    The cron can be managed in the future by the specified *label* 
    
    *Schedule* is a BSD-style cron timestamp, specified as either a string of 5 non-
    whitespace characters delimited by spaces or as a sequence (tuple, list, etc.) of 
    5 elements, each with non-whitespace characters.
    The ordering of the characters represent scheduling information for minutes, 
    hours, days, months, and days of week.                   
    See full format specification at http://unixhelp.ed.ac.uk/CGI/man-cgi?crontab+5        
    Year is not supported.  The N/S increment format is though.
                    
    PiCloud schedules your crons under the *GMT (UTC+0) timezone*.

    Certain special *kwargs* associated with cloud.call can be attached to the periodic jobs: 

    * _cores:
        Set number of cores your job will utilize. See http://docs.scivm.com/primer.html#choose-a-core-type/
        In addition to having access to more CPU cores, the amount of RAM available will grow linearly.
        Possible values for ``_cores`` depend on what ``_type`` you choose:
        
        * c1: 1                    
        * c2: 1, 2, 4, 8
        * f2: 1, 2, 4, 8, 16                                    
        * m1: 1, 2, 4, 8
        * s1: 1        
    * _env:
        A string specifying a custom environment you wish to run your job within.
        See environments overview at http://docs.scivm.com/environment.html                
    * _fast_serialization:
        This keyword can be used to speed up serialization, at the cost of some functionality.
        This affects the serialization of the spawned jobs' return value.
        The stored function will always be serialized by the enhanced serializer, with debugging features.
        Possible values keyword are:
                    
        0. default -- use cloud module's enhanced serialization and debugging info            
        1. no debug -- Disable all debugging features for result            
        2. use cPickle -- Use python's fast serializer, possibly causing PicklingErrors
    * _max_runtime:
        Specify the maximum amount of time (in integer minutes) jobs can run. If the job runs beyond 
        this time, it will be killed.                      
    * _os_env_vars:
        List of operating system environment variables that should be copied to PiCloud from your system
        Alternatively a dictionary mapping the environment variables to the desired values.
    * _priority: 
        A positive integer denoting the job's priority. PiCloud tries to run jobs 
        with lower priority numbers before jobs with higher priority numbers.            
    * _profile:
        Set this to True to enable profiling of your code. Profiling information is 
        valuable for debugging, but may slow down your jobs.
    * _restartable:
        In the very rare event of hardware failure, this flag indicates that a spawned 
        job can be restarted if the failure happened in the middle of it.
        By default, this is true. This should be unset if the function has external state
        (e.g. it modifies a database entry)
    * _type:
        Choose the type of core to use, specified as a string:
            
            * c1: 1 compute unit, 300 MB ram, low I/O (default)                    
            * c2: 2.5 compute units, 800 MB ram, medium I/O
            * f2: 5.5 compute units, 3.75 GB ram, high I/O, hyperthreaded core                                    
            * m1: 3.25 compute units, 8 GB ram, high I/O
            * s1: Up to 2 compute units (variable), 300 MB ram, low I/O, 1 IP per core
                       
        See http://www.scivm.com/pricing/ for pricing information
    * _vol:
        A string or list of strings specifying a volume(s) you wish your jobs to have access to.          
                 
    .. note:: 
    
        The callable must take no arguments.  To schedule functions that take arguments,
        wrap it with a closure. e.g. func(x,y) --> lambda: func(x,y)
    """
    
    cloud = _getcloud()
    params = cloud._getJobParameters(func,
                                             kwargs, 
                                             ignore=['_label', '_depends_on', '_depends_on_errors'])
    
    return _low_level_register(func, label, schedule, params)    

def deregister(label):
    """Deregister (delete) the cron specified by *label*"""

    #ASCII label:
    try:
        label = label.decode('ascii').encode('ascii')
    except (UnicodeDecodeError, UnicodeEncodeError):
        raise TypeError('label must be an ASCII string')
    
    conn = _getcloudnetconnection()
    
    conn.send_request(_deregister_query, {'label': label})   
        
def manual_run(label):
    """Manually run the function of a cron specified by *label* (ignoring schedule)
    Returns the 'jid' of the job created by this run command""" 

    #ASCII label:
    try:
        label = label.decode('ascii').encode('ascii')
    except (UnicodeDecodeError, UnicodeEncodeError):
        raise TypeError('label must be an ASCII string')
    
    conn = _getcloudnetconnection()
    
    resp = conn.send_request(_run_query, {'label': label})
    return resp['jid']   
    
def list():
    """
    List labels of active crons registered on PiCloud
    """ 
    
    conn = _getcloudnetconnection()
    resp = conn.send_request(_list_query, {})
    return resp['labels']

def info(label):
    """    
    Return a dictionary of relevant info about cron specified by *label*
    
    Info includes:
        
    * label: The same as was passed in
    * schedule: The schedule of this cron
    * last_run: Time this cron was last run
    * last_jid: Last run jid of this cron
    * created: When this cron was created   
    * creator_host: Hostname of the machine that registered this cron    
    * func_name: Name of function associated with cron
    """
    
    conn = _getcloudnetconnection()
    resp = conn.send_request(_info_query, {'label':label})
    del resp['data']
    #del resp['version']
    return resp    



"""
PiCloud account management.
This module allows the user to manage account information programmatically.
Currently supports api key management
"""
from __future__ import absolute_import
"""
Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2011 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see
http://www.gnu.org/licenses/lgpl-2.1.html
"""

import scicloud as cloud

_key_list = 'key/list/'
_key_get = 'key/%s/'
_key_activate = 'key/%s/activate/'
_key_deactivate = 'key/%s/deactivate/'
_key_create = 'key/'

def list_keys(username, password, active_only=False):
    """Returns a list of all api keys. If *active_only* is True, only
    active keys are returned. *username* and *password* should be your
    PiCloud login information."""

    conn = cloud._getcloudnetconnection()
    resp = conn.send_request(_key_list,
                             {},
                             get_values={'active_only': active_only},
                             auth=(username, password))

    return resp['api_keys']

def get_key(username, password, api_key):
    """Returns information including api_secretkey, active status, and
    note for the specified *api_key*. *username* and *password* should
    be your PiCloud login information."""

    conn = cloud._getcloudnetconnection()
    resp = conn.send_request(_key_get % api_key,
                             {},
                             auth=(username, password))

    return resp['key']



def activate_key(username, password, api_key):
    """Activates the specified *api_key*. *username* and *password*
    should be your PiCloud login information."""

    conn = cloud._getcloudnetconnection()
    resp = conn.send_request(_key_activate % api_key,
                             {},
                             auth=(username, password))

    return True

def deactivate_key(username, password, api_key):
    """Deactivates the specified *api_key*. *username* and *password*
    should be your PiCloud login information."""

    conn = cloud._getcloudnetconnection()
    resp = conn.send_request(_key_deactivate % api_key,
                             {},
                             auth=(username, password))

    return True

def create_key(username, password):
    """Creates a new api_key. *username* and *password*
    should be your PiCloud login information."""

    conn = cloud._getcloudnetconnection()
    resp = conn.send_request(_key_create,
                             {},
                             auth=(username, password))

    return resp['key']


def get_key_by_key(api_key, api_secretkey):
    """
    Similar to *get_key*, but access information via api_key credentials
    (api_key and api_secretkey).
    """

    conn = cloud._getcloudnetconnection()
    resp = conn.send_request(_key_get % api_key,
                             {},
                             auth=(api_key, api_secretkey))

    return resp['key']



"""
Science VM network connections
This module manages communication with the Science VM server

Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2010 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""
from __future__ import with_statement
import errno
import os
import sys
import time
import base64
import urllib
import httplib
import random
import socket
import urllib2
import threading
from datetime import datetime

from ..util import urllib2_file
from ..util import credentials
from ..cloud import CloudException
from .. import cloudconfig as cc
from .connection import CloudConnection

import logging
cloudLog = logging.getLogger('Cloud.HTTPConnection')

try:
    from json import dumps as serialize
    from json import loads as deserialize
    from json import JSONDecoder    
except ImportError: #If python version < 2.6, we need to use simplejson
    from simplejson import dumps as serialize
    from simplejson import loads as deserialize
    from simplejson import JSONDecoder
_decoder = JSONDecoder()    
    
#xrange serialization:
from ..util.xrange_helper import encode_maybe_xrange, decode_maybe_xrange, iterate_xrange_limit 

#zipping:
from ..util.zip_packer import Packer

#version transport:
from ..versioninfo import release_version

def unicode_container_to_str(data):
    """Recursively converts data from unicode to str.
    Responses from the server may be in unicode."""
    
    if isinstance(data, unicode):
        try:
            return str(data)
        except UnicodeEncodeError: # fall back to unicode
            return data
    elif isinstance(data, dict):
        return dict(map(unicode_container_to_str, data.iteritems()))
    elif isinstance(data, (list, tuple, set, frozenset)):
        return type(data)(map(unicode_container_to_str, data))
    else:
        return data
  

class HttpConnection(CloudConnection):
    """
    HTTPConnnection finds an available cloud cluster, and provides
    a gateway to query it.
    """
    
    api_key = cc.account_configurable('api_key',
                                     default='None',
                                     comment='your application\'s api key provided by PiCloud')
    api_secretkey = cc.account_configurable('api_secretkey',
                                           default='None',
                                           comment='secretkey. For backwards compatibility; replaced with credentials in 2.4',
                                           hidden = True)
    
    __api_default_url = 'http://api.scivm.com/servers/list/'
    server_list_url = cc.account_configurable('server_list_url',
                                           default=__api_default_url,
                                           comment="url to list of PiCloud servers",hidden=False) 
    #hack for users utilizing old api
    if server_list_url == 'http://www.scivm.com/pyapi/servers/list/':
        server_list_url = __api_default_url
    
    job_cache_size = cc.transport_configurable('cloud_status_cache_size',
                                  default=65536,
                                  comment="Number of job statuses to hold in local memory; set to 0 to disable caching. This option only applies when connected to PiCloud.")
    
    result_cache_size = cc.transport_configurable('cloud_result_cache_size',
                                  default=4096,
                                  comment="Amount (in kb) of job results to hold in local memory; set to 0 to disable caching. This option only applies when connected to PiCloud.")
    
    retry_attempts = cc.transport_configurable('retry_attempts',
                                               default=3, 
                                               comment='Number of times to retry requests if http error occurs',
                                               hidden=True)
    
    
    call_query = 'job/'
    map_query = 'job/map/add/'
    map_reduce_query = 'bigdata/map_reduce/add/'
    status_query =  'job/status/'
    result_query = 'job/result/serialized/'
    result_query_global = 'job/result/serialized_global/'
    kill_query= 'job/kill/'
    delete_query= 'job/delete/'    
    info_query = 'job/info/'
    modules_add_query = 'module/add/'
    modules_check_query = 'module/check/'
    packages_list_query = 'package/list/'
        
    # Most status codes are interpretted by the server
    status_accept = 200
    
    url = cc.account_configurable('url', default='',
                                comment="url to scivm server.  Set by server_list_url if not found",hidden=True)
    
    hostname = cc.transport_configurable('hostname',
                                  default='',
                                  comment="Internal use only: hardcodes hostname.", hidden = True)
    
    #used to track cloud graph for webview
    parent_jid = cc.transport_configurable('parent_jid',
                                  default=-1, #default must be an int for configs to work
                                  comment="Internal use only: Tracks cloud graph for webview", hidden = True)
    if parent_jid < 0: #flag for None.  
        parent_jid = None
    
    #auto packages version received from server (or set by PiCloud)
    __ap_version = cc.transport_configurable('ap_version', default='', 
                                        comment="Internal use only. Deals with module versioning",
                                        hidden = True)  
    
    
    api_key_validated = False # Set to true once key has been validated
    serverlist_resolved_url = False # Was url automatically resolved from server_list? 
    
    # protocol version
    # 1.0 --- xranges used to communicate
    # 1.1 --- new map uploading protocol
    # 1.2 --- GZIP individual elements
    # 1.3 --- Don't transport exception w/ status. Get through info
    # 2.0 --- New Cloud API
    # 2.1 --- Provide HTTP Error codes with responses
    
    version = '2.1'
    map_size_limit = 1800000 #1.8 MB map limit    
    map_job_limit = 500 #maximum number of map jobs per request    
    jid_size_limit = 10000 #maximum number of jids per request
    result_size_limit = 2000 # maximum number of results per request
        
               
    def __init__(self, api_key, api_secretkey=None, server_url=None):
        if api_key:
            self.api_key = str(api_key)
        elif os.environ.get('PICLOUD_API_KEY',None):
            self.api_key = str(os.environ.get('PICLOUD_API_KEY'))
        if api_secretkey:
            self.api_secretkey = str(api_secretkey)
        elif os.environ.get('PICLOUD_API_SECRETKEY',None):
            self.api_secretkey = str(os.environ.get('PICLOUD_API_SECRETKEY'))
        else:  #resolve it later when sent
            self.api_secretkey = None
            
        if server_url:
            self.url= server_url
        if not self.hostname:
            self.hostname = str(socket.gethostname())
        self.openLock = threading.RLock()
        
        #module caching:
        self._modsInspected = set()
        self._modVersions = {}

    """Module version analysis"""
    def _get_mod_versions(self):
        for modname, module in sys.modules.items():
            if modname in self._modsInspected:
                continue            
            self._modsInspected.add(modname)
            if '.' in modname:  #only inspect top level
                continue      
            version_strings = ['__version__', 'version', 'Version']      
            for vs in version_strings:
                val = getattr(module, vs, None)    
                if isinstance(val, (str, int, long)): #must be primitive to be JSON'd
                    self._modVersions[modname] = val
        return self._modVersions            
        
    def diagnose_network(self):
        """Verify if user's network connection is online"""
        
        general_addr = 'http://www.google.com'
        try:
            self.post(general_addr)
        except Exception, e:
            raise CloudException('Cannot connect to PiCloud. Diagnostic indicate that your computer cannot connect to %s. Your network is likely down.' % general_addr)

        aws_addr = 'http://aws.amazon.com'
        try:
            self.post(aws_addr)
        except Exception, e:
            raise CloudException("Cannot connect to PiCloud. Diagnostic indicate that PiCloud's provider, Amazon Webservices are down. Please check http://status.aws.amazon.com/")
        
    
    def resolve_by_serverlist(self):
        self.serverlist_resolved_url = True

        try:
            # Must not call send_request as that can re-enter this function on failure
            resp = self.send_request_helper(self.server_list_url, {})
        except CloudException:
            self.diagnose_network() # raises diagnostic exception if diagnosis fails
            raise # otherwise something wrong with PiCloud
        
        for accesspoint in resp['servers']:
            try:
                cloudLog.debug('Trying %s' % accesspoint)
                # see if we can connect to the server
                req = urllib2.Request(accesspoint)                    
                resp = urllib2_file.urlopen(req, timeout = 30.0)
                resp.read()
            except Exception:
                cloudLog.info('Could not connect to %s', exc_info = True)
                pass                        
            else:
                self.url = str(accesspoint)
                cloudLog.info('Connected to %s' % accesspoint)
                if req.get_type() != 'https':
                    cloudLog.warning('Connected over an insecure connection. Be sure that openssl and python-openssl are installed')
                break
        else:
            # if it could not establish a connection any of the listed servers
            raise CloudException('HttpConnection.__init__: Could not find working cloud server at %s. Please contact PiCloud about this error' % datetime.utcnow(),
                                 logger=cloudLog, status=503)        
        
    def open(self, force_open=True):
        """open adapter
        force_open is used to force resolution of PiCloud url even if no api key specified
        """                           
        with self.openLock:
            
            if self.opened:  #ignore multiple opens
                return
            
            if self.api_key == 'None' and not force_open:
                cloudLog.debug('No api_key set: using dummy connection')
                self.url = ''
                return False
            
            if self.adapter:
                #Cache configuration copying.
                #0 implies no cache (which internally is None)                    
                self.adapter.cloud.job_cache_size = self.job_cache_size if self.job_cache_size > 0 else None                
                self.adapter.cloud.result_cache_size = self.result_cache_size*1024 \
                    if self.result_cache_size > 0 else None
                    
                if not self.adapter.opened:
                    self.adapter.open()
            
            # get list of available servers if no url
            if not self.url:
                self.resolve_by_serverlist()
            
            #finish open
            self._isopen = True

    def connection_info(self):
        dict = CloudConnection.connection_info(self)
        dict['connection_type'] = 'HTTPS' if 'https://' in self.url else 'HTTP'
        dict['server_url'] = self.url
        dict['api_key'] = self.api_key
        dict['api_secretkey'] = self.api_secretkey
        return dict
    
    def needs_restart(self, **kwargs):
        
        api_key = kwargs.get('api_key')
        if api_key:
            if api_key != self.api_key:
                return True
        server_url = kwargs.get('server_url')
        if server_url:
            if server_url != self.url:
                return True        
        return False
    
    def get_secretkey(self):
        """Return api_secretkey associated with current api_key"""
        
        if self.api_secretkey:  #cached or manually set
            api_secretkey = self.api_secretkey
        else:
            api_secretkey = credentials.resolve_secretkey(self.api_key) #None if not resolved
            if not api_secretkey: #fall back to secretkey in cloudconf (backwards compatibility)
                api_secretkey = self.__class__.api_secretkey
            if not api_secretkey or api_secretkey == 'None':
                raise CloudException('HttpConnection: api_secretkey for key %s not found. Please run scivm setup' % self.api_key)
        
        if not self.api_key_validated:
            #Download key if any components are missing
            if not credentials.verify_key(self.api_key):
                credentials.download_key_by_key(self.api_key, api_secretkey)
            self.api_key_validated = True
        
        self.api_secretkey = api_secretkey
        return api_secretkey
    
    def post(self, url, post_values=None, headers={}, use_gzip=False):
        """Simple HTTP POST to the input url with the input post_values.
        If post_values are None, HTTP get issued"""
        request = urllib2.Request(url, post_values, headers)
        request.use_gzip = use_gzip
        response = urllib2_file.urlopen(request)
        
        return response
    
    def send_request_helper(self, url, post_values=None, headers={}, raw_response=False, log_cloud_excp = True):
        """Creates an http connection to the given url with the post_values encoded.
        Returns the http response if the returned status code is > 200."""
        
        post_params = [] if post_values != None else None
        #post_values['version'] = self.version
        
        # remove None values
        if post_values:
            for key in post_values.keys():
                if post_values[key] == None:
                    del post_values[key]
                elif isinstance(post_values[key], (tuple, list)):
                    for v in post_values[key]:
                        post_params.append((key, v))
                    del post_values[key]
            
            post_params.extend(post_values.items()) 
        
        attempt = 0

        while attempt <= self.retry_attempts:
            try:
                body = None
                
                try:
                    response = self.post(url, post_params, headers)
                except urllib2.HTTPError, e:
                    # 4xx errors have our error_codes in their body
                    # 520 is special "unexpected scivm server error code"
                    
                    if 400 <= e.code < 500 or e.code == 520:  
                        response = e
                    elif 200 < e.code < 300: # python 2.5 bug
                        response = e 
                    else:
                        raise                        
                
                # read entire response
                body = response.read()
                response.close()
               
                if raw_response:
                    resp = body
                    break
                else:                    
                    resp = self.parse_response(body)
                    error = resp.get('error')
                    if error:
                        raise CloudException(error['msg'], status=error['code'], 
                                         retry=error['retry'], logger=cloudLog)
                    
                    elif response.code > 300:  # 4xx error without an error_code in body; somnething is wrong 
                        raise response
                    else:
                        break
            
            except Exception, e:
                
                # only retry data transfer errors
                if not isinstance(e, (IOError, httplib.HTTPException, socket.error, CloudException)):
                    raise                
                
                attempt += 1
                
                # did we get an exception that should always be raised?
                must_raise_exc = isinstance(e, CloudException) and not e.retry
                                
                # was this a connection refused error?
                conn_refused = isinstance(e, socket.error) and getattr(e, 'errno', e.args[0]) == errno.ECONNREFUSED or \
                    (isinstance(e, urllib2.HTTPError) and e.code in [500, 503])                   
                
                if not must_raise_exc:                    
                    if attempt <= self.retry_attempts:
                        cloudLog.warn('rawquery: Problem connecting to PiCloud. Retrying. \nError is %s' % str(e))
                        logfunc = cloudLog.warn    
                    else:
                        cloudLog.exception('rawquery: http connection failed')
                        logfunc = cloudLog.error
                    
                    # HTTP Error
                    if getattr(e, 'readlines', None):
                        for line in e.readlines():
                            logfunc(line.rstrip())
                    
                    # if we received a body that we could not parse successfully
                    if body and not isinstance(e, CloudException):
                        if isinstance(e, urllib2.HTTPError):
                            err_code = e.code
                        else:
                            err_code = '2xx'
                        logfunc('Specifically HTTP %s with invalid data.' % err_code)
                        for line in body.split('\n'):
                            logfunc(line.strip())
                
                # log error details
                if attempt > self.retry_attempts or must_raise_exc:
                    if isinstance(e, CloudException):
                        if log_cloud_excp:
                            cloudLog.error('rawquery: received error from server: %s', e)
                        raise
                    else:
                        cloudLog.exception('rawquery: PiCloud appears to be unavailable. Showing traceback')
                        raise CloudException('PiCloud is unavailable. Please contact PiCloud with this message. Specific error:  %s - %s' % (datetime.utcnow(), e), 
                                             status=503)
                else:
                    # exponential backoff
                    c = attempt -1 
                    if conn_refused:
                        # on connection refused, use a higher base sleep (5 seconds)
                        time.sleep(min(60, 5 + (1 << c) *random.random()))
                    else:
                        time.sleep(min(30, (1 << c) *random.random()))
                    continue
        
        return resp
    
    
    def parse_response(self, body):
        """*body* is a JSON-encoded string of a dictionary, that may
        or may not have raw binary data appended after it. The return
        value will be a dictionary of the loaded JSON object, and if
        *body* included raw binary data, that data will be accessible
        from the 'data' key."""
        
        try:
            obj, end = _decoder.raw_decode(body)
        except ValueError, v:
            # handle no json object could be decoded        
            raise CloudException('Could not process response due to error %s. Are you behind a firewall? \nServer unexpectedly returned %s' %  \
                                 (str(v), body))
        
        resp = obj
        resp['data'] = body[end:]
        
        return resp
        
    
    def send_request(self, url, post_values, get_values=None, logfunc=cloudLog.info, raw_response=False, log_cloud_excp = True, auth=None):
        """High-level HTTP sending logic
        url -- url to communicate with
        post_values -- dictionary defining postdata
        get_values -- dictionary of items defining getdata (appended to url)
        logfunc -- function to be used for logging
        raw_response -- If true, do not parse body with 'parse_response'
        log_cloud_excp -- If a cloudexception is received from server, should it be logged with logging.error
        auth -- 2 element tuple defining username, password -- overrides api_key/api_secretkey if provided
        """  
        
        if not auth and self.api_key == 'None':
            raise CloudException('HttpConnection.query: api_key is not set. Please run "scivm setup" or call cloud.setkey()', logger=cloudLog)
                
        headers = {}
        
        if auth:
            base64string = base64.encodestring('%s:%s' % auth)[:-1]
        else:
            api_secretkey = self.get_secretkey()
            base64string = base64.encodestring('%s:%s' % (self.api_key, api_secretkey))[:-1]
        base64string = base64string.replace('\n','') # remove newlines (base64 generates them every 76th char)
        
        if logfunc:
            logfunc('query url %s with post_values =%s' % (url, post_values))        
        
        headers['Authorization'] = 'Basic %s' % base64string

        # add general information
        #post_values.update({'api_key': self.api_key,
        #                    'api_secretkey': self.api_secretkey})
        get_values = get_values or {}
        get_values['version'] = self.version
        
        # urlencode dict with sequence values
        query_string = urllib.urlencode(get_values, True) if get_values else ''
        if query_string:
            url += '?' + query_string

        if not url.startswith('http://') and not url.startswith('https://'):
            url = self.url + url 
        
        try:
            return self.send_request_helper(url, post_values, headers, raw_response = raw_response, 
                                        log_cloud_excp = log_cloud_excp)
        except CloudException, ce:            
            if ce.status == 503: # network error
                if self.serverlist_resolved_url:
                    self.resolve_by_serverlist() # re-resolve server list (or raise network diagnostic error)
                    # try again:
                    return self.send_request_helper(url, post_values, headers, raw_response = raw_response, 
                                        log_cloud_excp = log_cloud_excp)
                else:
                    # try diagnosing network; raises custom exception if diag fails; otherwise continue on.
                    self.diagnose_network() 
            
            # cannot resolve problem or identify a different error
            raise 
    
    def is_simulated(self):
        return False
    
    def _update_params(self, params):

        params.update({'hostname': str(self.hostname),
                       'process_id': str(os.getpid()),                       
                       'language_version': sys.version,
                       'ap_version': self.__ap_version,
                       'parent_jid' : self.parent_jid
                       })
                
        params.setdefault('language', 'python') # allow parent to declare a different lang (e.g. shell)
        params['mod_versions'] = self._get_mod_versions() # module versions for tracking
        params['cloud_version'] = release_version        
        
    
    def job_add(self, params, logdata=None):
        
        # let adapter make any needed calls for dep tracking
        self.adapter.dep_snapshot()
        
        self._update_params(params)
        
        #strip unicode from func_name:
        params['func_name'] = params['func_name'].decode('ascii', 'replace').encode('ascii', 'replace')
        
        data = Packer()
        
        data.add(params['func'])
        data.add(params['args'])
        data.add(params['kwargs'])
        
        del params['func'], params['args'], params['kwargs']
        
        params['data'] = data.finish()
        
        if params['depends_on']:
            params['depends_on'] = self.pack_jids(params['depends_on'])[0]

        resp = self.send_request(self.call_query, params, logfunc=cloudLog.debug)
        
        jid = resp['jids']
        
        cloudLog.info('call %s --> jid [%s]', params['func_name'], jid)
        
        return jid
    
    def jobs_map(self, params, mapargs, mapkwargs = None, logdata=None):
        self._update_params(params)
        
        if not mapargs and not mapkwargs:
            raise ValueError('either mapargs or mapkwargs must be non-None')
                
        #strip unicode from func_name:
        params['func_name'] = params['func_name'].decode('ascii', 'replace').encode('ascii', 'replace')
        
        data = Packer()
        
        data.add(params['func'])
        size = len(params['func'])
        
        if params['depends_on']:
            params['depends_on'] = self.pack_jids(params['depends_on'])[0]        
        
        del params['func']
        
        # done indicates to the server when the last chunk of the map is being sent
        params['done'] = False
        
        # this tells the server what the first maparg index of the current map chunk is
        # for ex. first_maparg_index=0 (first request with 4 chunks),
        #         first_maparg_index=4 (second request)
        cnt = 0
        req_item_cnt = 0
        params['first_maparg_index'] = cnt
        
        first_iteration = True
        map_is_done = False
        
        argIter =  iter(mapargs) if mapargs else None
        kwdIter = iter(mapkwargs) if mapkwargs else None
        
        fname = params['func_name']
        
        next_elm = None
        next_kwd = None
        
        if kwdIter:
            kwd_data = Packer()
        else:
            kwd_data = None
        
        while True:

            try:
                if argIter:                
                    next_elm = argIter.next()                
                if kwdIter:
                    next_kwd = kwdIter.next()                                
            except StopIteration:
                map_is_done = True
                next_elm = None
                next_kwd = None
                params['done'] = True                
            
            if (cnt and size > self.map_size_limit) or req_item_cnt > self.map_job_limit or map_is_done:
                
                if cnt == 0:  #empty mapargs - don't send anything
                    return []
                
                self.adapter.dep_snapshot() #let adapter make any needed calls for dep tracking
                if data:
                    params['data'] = data.finish()  #payload
                if kwd_data:
                    params['map_kwargs'] = kwd_data.finish()
                params['hostname'] = str(self.hostname)
                params['ap_version'] = self.__ap_version
                
                resp = self.send_request(self.map_query, params, logfunc=cloudLog.debug) #actual query
                
                if not map_is_done:

                    if first_iteration:
                        # extract group_id from the response
                        group_id = resp['group_id']                                        
                    
                    # reset parameters
                    params = {'group_id': group_id,
                              'done': False,
                              'first_maparg_index': cnt}
                    
                    #rebuild data object                    
                    data = Packer() if argIter else None
                    if kwdIter:
                        kwd_data = Packer()
                        data = Packer()
                    else:
                        kwd_data = None
                    
                    
                else:
                    break
                
                size = 0
                req_item_cnt = 0
                
                # set first iteration to false only after the first *map chunk is sent*
                first_iteration = False

            if next_elm:
                data.add(next_elm)
                size += len(next_elm) 
                cnt += 1
                req_item_cnt += 1
            if next_kwd:
                kwd_data.add(next_kwd)
                size += len(next_kwd) 
                cnt += 1
                req_item_cnt += 1
            
        jids = decode_maybe_xrange(resp['jids'])
        
        cloudLog.info('map %s --> jids [%s]', fname, jids)
        
        return jids
    
    def add_map_reduce_job(self, params):
        
        self.adapter.dep_snapshot()
        
        self._update_params(params)
        
        data = Packer()
        
        data.add(params['mapper_func'])
        data.add(params['reducer_func'])
        data.add(params['bigdata_file'])
        
        del params['mapper_func'], params['reducer_func'], params['bigdata_file']
        
        params['data'] = data.finish()
        
        resp = self.send_request(self.map_reduce_query, params, logfunc=cloudLog.debug)
        
        jid = resp['jids']
        
        return jid
        
    
    @staticmethod
    def pack_jids(jids):
        packedJids = Packer()
        serialized_jids = serialize(encode_maybe_xrange(jids))
        packedJids.add(serialized_jids)
        return packedJids.finish(), serialized_jids
    
    def jobs_result(self, jids, by_jid):
        # resp is in a pseudo-multipart format with a boundary
        # separating result fields

        results = {}
        for rjids in iterate_xrange_limit(jids,self.result_size_limit):    

            packed_jids, serialized_jids = self.pack_jids(rjids)
            if not by_jid:
                cloudLog.info('query result of jids %s' % serialized_jids) 

            if by_jid:
                resp = self.send_request(self.result_query_global, {'jids': packed_jids}, logfunc=None)
            else:
                resp = self.send_request(self.result_query, {'jids': packed_jids}, logfunc=None)
            
            # THINK: We could throw a warning if there is potentially a Python version incompatibility
            if resp['language'] != 'python':
                raise CloudException('HttpConnection.jobs_result: Result data is not Python compatible because it was generated in %s.' % resp['language'], logger=cloudLog)
            
            data = resp['data']
            
            boundary_index = data.index('\n')
            
            # parse boundary definition
            boundary = data[len('boundary='):boundary_index]
            
            # separate the boundary definition from the rest of the data
            rest = data[boundary_index+1:]
            
            # split results by boundary
            results_data = rest.split(boundary)
            interpretation = resp['interpretation']
            # filter out empty strings
            results.setdefault('data',[]).extend( [datum for datum in results_data if datum] )
            results.setdefault('interpretation',[]).extend(interpretation)
            
        return results
        
    def jobs_kill(self, jids):                
        if jids == None:
            #send 'kill all' command to server, which is encoded as kill([])
            packed_jids = self.pack_jids([])[0]
            cloudLog.info('kill all jobs')
            self.send_request(self.kill_query, {'jids': packed_jids}, logfunc=None)
            
        for rjids in iterate_xrange_limit(jids,self.jid_size_limit):
            packed_jids, serialized_jids = self.pack_jids(rjids)
            cloudLog.info('kill jids %s' % serialized_jids)
            self.send_request(self.kill_query, {'jids': packed_jids}, logfunc=None)

    def jobs_delete(self, jids):
        for rjids in iterate_xrange_limit(jids,self.jid_size_limit):
            packed_jids, serialized_jids = self.pack_jids(rjids)
            cloudLog.info('delete jids %s' % serialized_jids)
            self.send_request(self.delete_query, {'jids':  packed_jids}, logfunc=None)
        
    def jobs_info(self, jids, info_requested):
        
        infos = {}
        for rjids in iterate_xrange_limit(jids, self.jid_size_limit):    
            
            packed_jids, serialized_jids = self.pack_jids(rjids)        
            serialized_info = serialize(info_requested)
            cloudLog.info('query [%s] on jids %s' % (serialized_info, serialized_jids)) 
            
            resp = self.send_request(self.info_query,
                                     post_values={'jids':  packed_jids},
                                     get_values={'field': info_requested},
                                     logfunc=None)
            info_dct = resp['info']
                        
            for (x,y) in info_dct.iteritems():
                try:
                    x = long(x)
                except ValueError:
                    x = str(x)
                y = unicode_container_to_str(y)     
                
                # additional patching
                for k in y.keys():                
                    if k in ['func_object', 'args', 'kwargs']:
                        y[k] = base64.b64decode(y[k])
                
                infos[x] = y
            
        return infos
    
    def modules_check(self, modules):
        """modules_check determines which modules must be sent from the client
        to the server.
        modules: list of tuples where each tuple is (filename, timestamp)
        
        Returns a list of filenames to send."""
        
        packedMods = Packer()
        packedMods.add(serialize(modules))      
        data = packedMods.finish()

        resp = self.send_request(self.modules_check_query,
                                 {'data': data,
                                  'hostname': str(self.hostname),
                                  'language': 'python'})        
        
        mods = resp['modules']
        
        if 'ap_version' in resp:
            self.__ap_version = resp['ap_version']
            #cloudLog.info("network.py: modules_check(): result['ap_version'] of query: %s" % resp['ap_version'])
            
        cloudLog.info('network.py: modules_check(): ap_version is now %s. needed mods are %s', 
                      self.__ap_version, mods)
        
        return mods
    
    def modules_add(self, modules, modules_tarball):
        """modules_add adds the specified modules to the scivm system.
        modules is a list of tuples, where each tuple is (name, timestamp).
        modules_tarball is a string representing the tarball of all the included modules."""
        
        packedMods = Packer()
        packedMods.add(serialize(modules))
        packedMods.add(modules_tarball)        
        data = packedMods.finish()

        resp = self.send_request(self.modules_add_query,
                                 {'data': data,
                                  'hostname': str(self.hostname),
                                  'language': 'python'})
        
        if 'ap_version' in resp:
            self.__ap_version = resp['ap_version']
            #cloudLog.info("network.py: modules_add(): result['ap_version'] of query: %s" % resp['ap_version'])
            
        cloudLog.info('network.py: modules_add(): ap_version is %s' % self.__ap_version)
    
    
    def report_install(self):
        
        pass
        
    def packages_list(self):
        """Get list of pre-installed packages from server"""
        
        resp = self.send_request(self.packages_list_query,
                                 {'language': 'python',
                                  'language_version': sys.version})
        
        # convert from unicode to ascii
        return map(str, resp['packages'])
    
    def report_name(self):
        return 'HTTPConnection'
        
    



"""
This module provides functions to "wait" on various aspects of a job

All functions expect a single jid.
A timeout may also be specified.

Return value/exception handling is function specific
"""
""" 

Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2012 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

import time as _time

from .cloud import CloudException as _CloudException
from .cloud import CloudTimeoutError as _CloudTimeoutError
from . import _getcloud


def _wait_for_test(jid, test_func, timeout=None, timeout_msg='Job wait timed out'):
    """Keep testing job until test_func(jid) returns a true value
    Return return value of test_func"""    
    
    poll_interval = 1.0    
    abort_time = _time.time() + timeout if timeout else None  

    while True:
        retval = test_func(jid)
        if retval:
            return retval
        
        if abort_time and _time.time() > abort_time:
            raise _CloudTimeoutError(timeout_msg, jid=jid)
             
        _time.sleep(poll_interval)        

def _checkint(var, name=''):
    if not isinstance(var, (int, long)):
        raise TypeError('%s must be a single integer' % name)


# all possible status transitions
_status_transitions = {'waiting' : ['stalled', 'killed', 'queued'],
                       'queued' : ['killed', 'processing'],
                       'processing' : ['killed', 'error', 'done'],
                       'stalled' : [],
                       'killed' : [],
                       'error' : [],
                       'done' : []
                       }

# all possible future states (could autogenerate from above)
_possible_future_statuses = {'waiting' : ['stalled', 'killed', 'queued', 'processing', 'done', 'error'],
                             'queued' : ['killed', 'processing', 'done', 'error'],
                             'processing' : ['killed', 'error', 'done'],
                             'stalled' : [],
                             'killed' : [],
                             'error' : [],
                             'done' : []
                             }

def _status_test_wrapper(test_status):
    """wrapper function to conduct status tests"""
    return status_test

def status(jid, test_status, timeout=None):
    """Wait until job's status is ``test_status``
    Raise CloudException if no longer possible to reach status
    
    Returns job's current status (which will be equal to test_status)    
    """
    _checkint(jid, 'jid')
    def status_test(jid):
        cl = _getcloud()
        cur_status = cl.status(jid)
        if test_status == cur_status:
            return cur_status
        if test_status not in _possible_future_statuses[cur_status]:
            raise _CloudException('Job has status %s. Will never (again) be %s' % (cur_status, test_status), 
                                 jid=jid, status=cur_status) 
        return False   
    
    return _wait_for_test(jid, status_test, timeout=timeout, 
                          timeout_msg='Job did not reach status %s before timeout' % test_status)

    
def port(jid, port, protocol='tcp', timeout=None):
    """Wait until job has opened ``port`` (under protocol ``protocol``) 
    for listening.
    Returns port translation dictionary. 
    
    See docstring for :func:`cloud.shortcuts.get_connection_info` for description
    of returned dictionary
    """
    
    _checkint(jid, 'jid')
    _checkint(port, 'port')

    cl = _getcloud()
        
    processing_poll_interval = 1.0 # polling on status wait
    port_poll_interval = 0.7 # polling on port wait
    
    abort_time = _time.time() + timeout if timeout else None
    status = None

    while True:
        jid_info = cl.info(jid, ['ports','status'])[jid]
        status = jid_info['status']
        port_info = jid_info.get('ports')
        
        if status in _possible_future_statuses['processing']:
            raise _CloudException('Job is already finished with status %s' % status, 
                        jid=jid, status=status)
        elif not port_info:
            if cl.is_simulated() and status == 'processing':
                return {'address' : '127.0.0.1', 'port' : port}
            elif abort_time and _time.time() > abort_time:
                raise _CloudTimeoutError('Job did not start processing before timeout', 
                                    jid=jid, status=status)            
            _time.sleep(processing_poll_interval)
            continue
        
        port_proto_info = port_info[protocol]        
        if port not in port_proto_info:
            if abort_time and _time.time() > abort_time:
                raise _CloudTimeoutError('Job did not open port %s before timeout' % port, 
                                    jid=jid)
            _time.sleep(port_poll_interval)
            continue
        
        return port_proto_info[port]           
    



'''
Cloud Multiprocessing Interface

cloud.mp has effectively the same interface as cloud.

All jobs will be run locally, on multiple processors, via the multiprocessing library

Sample usage::

    import scicloud.mp
    jid = scicloud.mp.call(lambda: 3*3)
    >> Returns a job identifier
    scicloud.mp.result(jid)
    >> Returns 9
'''
'''
Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2010 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see
http://www.gnu.org/licenses/lgpl-2.1.html
'''

import sys

import cloudinterface
from . import cloud

__cloud = None
__type = None
__immutable = False

#function bindings
call = None
map = None
status = None
join = None
result = None
kill = None
connection_info = None

__all__ = ["call", "map", "status", "result", "iresult", "join", "kill", "info",
           "delete", "connection_info", "finished_statuses", "close","c1","c2","m1"]

def _launch_cloud():
    cloudinterface._setcloud(sys.modules[__name__], 'mp')

_launch_cloud()

def _getcloud():
    """Return internal cloud object. Only for internal use"""
    return __cloud



"""
BSD-Style CronTab parsing
"""
"""
Code derived from 
http://android-scripting.googlecode.com/hg/python/gdata/samples/oauth/oauth_on_appengine/appengine_utilities/cron.py

Original notice:

Copyright (c) 2008, appengine-utilities project
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
- Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
- Neither the name of the appengine-utilities project nor the names of its
  contributors may be used to endorse or promote products derived from this
  software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE 
"""

"""
Possible to-do: Add 'next-run' support - what is next time cron will run?
"""

import os
import re
import datetime

class CronTime(object):
    """
    Parse a BSD-style cron timestamp and test if it should be run at a desired time
    See format at http://unixhelp.ed.ac.uk/CGI/man-cgi?crontab+5
    Also supports N/S increment format (e.g. 1/3 for every 3 min starting at min 1), as
        described at http://en.wikipedia.org/wiki/CRON_expression
    
    """

    def __init__(self, cron_string):
        cron = cron_string.split(" ")
        if len(cron) is not 5:
            raise ValueError, 'Invalid cron string. Format: min hour day mon dow'
        self.cron_raw = {
            'min': cron[0],
            'hour': cron[1],
            'day': cron[2],
            'mon': cron[3],
            'dow': cron[4],
        }
        self.cron_compiled = self._validate_cron(self.cron_raw)

    def _validate_cron(self, cron):
        """
        Parse the field to determine whether it is an integer or lists,
        also converting strings to integers where necessary. If passed bad
        values, raises a ValueError.
        """
        parsers = {
            'dow': self._validate_dow,
            'mon': self._validate_mon,
            'day': self._validate_day,
            'hour': self._validate_hour,
            'min': self._validate_min,
        }
        for el in cron:
            parse = parsers[el]
            cron[el] = parse(cron[el])
        return cron
        
    def _get_max(self, t):
        """
        Get maximum number for a given type
        """
        max_num = {'dow': 7,
             'mon': 12,
             'day':31,
             'hour':23,
             'min': 59
             }
          
        return max_num[t]

    def _validate_type(self, v, t):
        """
        Validates that the number (v) passed is in the correct range for the
        type (t). Raise ValueError, if validation fails.

        Valid ranges:
        day of week = 0-7
        month = 1-12
        day = 1-31
        hour = 0-23
        minute = 0-59

        All can * which will then return the range for that entire type.
        """
        if t == "dow":
            if v >= 0 and v <= 7:
                return [v]
            elif v == "*":
                return "*"
            else:
                raise ValueError, "Invalid day of week."
        elif t == "mon":
            if v >= 1 and v <= 12:
                return [v]
            elif v == "*":
                return range(1, 12)
            else:
                raise ValueError, "Invalid month."
        elif t == "day":
            if v >= 1 and v <= 31:
                return [v]
            elif v == "*":
                return range(1, 31)
            else:
                raise ValueError, "Invalid day."
        elif t == "hour":
            if v >= 0 and v <= 23:
                return [v]
            elif v == "*":
                return range(0, 23)
            else:
                raise ValueError, "Invalid hour."
        elif t == "min":
            if v >= 0 and v <= 59:
                return [v]
            elif v == "*":
                return range(0, 59)
            else:
                raise ValueError, "Invalid minute."

    def _validate_list(self, l, t):
        """
        Validates a crontab list. Lists are numerical values seperated
        by a comma with no spaces. Ex: 0,5,10,15

        Arguments:
            l: comma seperated list of numbers
            t: type used for validation, valid values are
                dow, mon, day, hour, min
        """
        elements = l.split(",")
        return_list = []
        # we have a list, validate all of them
        for e in elements:
            if "-" in e:
                return_list.extend(self._validate_range(e, t))
            else:
                try:
                    v = int(e)
                    self._validate_type(v, t)
                    return_list.append(v)
                except:
                    raise ValueError, "Names are not allowed in lists."
        # return a list of integers
        return return_list

    def _validate_range(self, r, t):
        """
        Validates a crontab range. Ranges are 2 numerical values seperated
        by a dash with no spaces. Ex: 0-10

        Arguments:
            r: dash seperated list of 2 numbers
            t: type used for validation, valid values are
                dow, mon, day, hour, min
        """
        elements = r.split('-')
        # a range should be 2 elements
        if len(elements) is not 2:
            raise ValueError, "Invalid range passed: " + str(r)
        # validate the minimum and maximum are valid for the type
        for e in elements:
            self._validate_type(int(e), t)
        # return a list of the numbers in the range.
        # +1 makes sure the end point is included in the return value
        return range(int(elements[0]), int(elements[1]) + 1)

    def _validate_step(self, s, t):
        """
        Validates a crontab step. Steps are complicated. They can
        be based on a range 1-10/2 or just step through all valid
        */2 or 1/2. When parsing times you should always check for step first
        and see if it has a range or not, before checking for ranges because
        this will handle steps of ranges returning the final list. Steps
        of lists is not supported.

        Arguments:
            s: slash seperated string
            t: type used for validation, valid values are
                dow, mon, day, hour, min
        """
        elements = s.split('/')
        # a range should be 2 elements
        if len(elements) is not 2:
            raise ValueError, "Invalid step passed: " + str(s)
        try:
            step = int(elements[1])
        except:
            raise ValueError, "Invalid step provided " + str(s)
        r_list = []
        # if the first element is *, use all valid numbers
        if elements[0] is "*" or elements[0] is "":
            r_list.extend(self._validate_type('*', t))
        # check and see if there is a list of ranges
        elif "," in elements[0]:
            ranges = elements[0].split(",")
            for r in ranges:
                # if it's a range, we need to manage that
                if "-" in r:
                    r_list.extend(self._validate_range(r, t))
                else:
                    try:
                        r_list.extend(int(r))
                    except:
                        raise ValueError, "Invalid step provided " + str(s)
        elif "-" in elements[0]:
            r_list.extend(self._validate_range(elements[0], t))
        else:   #picloud: support single number
            base = int(elements[0]) % step
            if base == 0 and t in ['mon','day']:
                base += step
            r_list.extend(self._validate_range('%d-%d' % (base, self._get_max(t)), t))        
        return range(r_list[0], r_list[-1] + 1, step)

    def _validate_dow(self, dow):
        """
        """
        # if dow is * return it. This is for date parsing where * does not mean
        # every day for crontab entries.
        if dow is "*":
            return dow
        if dow is '?':
            return '*' #convert '?' to '*'
        days = {
        'mon': 1,
        'tue': 2,
        'wed': 3,
        'thu': 4,
        'fri': 5,
        'sat': 6,
        # per man crontab sunday can be 0 or 7.
        'sun': [0, 7],
        }
        if dow in days:
            dow = days[dow]
            return [dow]
        # if dow is * return it. This is for date parsing where * does not mean
        # every day for crontab entries.
        elif dow is "*":
            return dow
        elif "/" in dow:
            return(self._validate_step(dow, "dow"))
        elif "," in dow:
            return(self._validate_list(dow, "dow"))
        elif "-" in dow:
            return(self._validate_range(dow, "dow"))
        else:
            valid_numbers = range(0, 8)
            if not int(dow) in valid_numbers:
                raise ValueError, "Invalid day of week " + str(dow)
            else:
                return [int(dow)]

    def _validate_mon(self, mon):
        months = {
        'jan': 1,
        'feb': 2,
        'mar': 3,
        'apr': 4,
        'may': 5,
        'jun': 6,
        'jul': 7,
        'aug': 8,
        'sep': 9,
        'oct': 10,
        'nov': 11,
        'dec': 12,
        }
        if mon in months:
            mon = months[mon]
            return [mon]
        elif mon is "*" or mon is '?':
            return range(1, 13)
        elif "/" in mon:
            return(self._validate_step(mon, "mon"))
        elif "," in mon:
            return(self._validate_list(mon, "mon"))
        elif "-" in mon:
            return(self._validate_range(mon, "mon"))
        else:
            valid_numbers = range(1, 13)
            if not int(mon) in valid_numbers:
                raise ValueError, "Invalid month " + str(mon)
            else:
                return [int(mon)]

    def _validate_day(self, day):
        if day is "*":
            return range(1, 32)
        elif "/" in day:
            return(self._validate_step(day, "day"))
        elif "," in day:
            return(self._validate_list(day, "day"))
        elif "-" in day:
            return(self._validate_range(day, "day"))
        else:
            valid_numbers = range(1, 31)
            if not int(day) in valid_numbers:
                raise ValueError, "Invalid day " + str(day)
            else:
                return [int(day)]

    def _validate_hour(self, hour):
        if hour is "*":
            return range(0, 24)
        elif "/" in hour:
            return(self._validate_step(hour, "hour"))
        elif "," in hour:
            return(self._validate_list(hour, "hour"))
        elif "-" in hour:
            return(self._validate_range(hour, "hour"))
        else:
            valid_numbers = range(0, 23)
            if not int(hour) in valid_numbers:
                raise ValueError, "Invalid hour " + str(hour)
            else:
                return [int(hour)]

    def _validate_min(self, min):
        if min is "*":
            return range(0, 60)
        elif "/" in min:
            return(self._validate_step(min, "min"))
        elif "," in min:
            return(self._validate_list(min, "min"))
        elif "-" in min:
            return(self._validate_range(min, "min"))
        else:
            valid_numbers = range(0, 59)
            if not int(min) in valid_numbers:
                raise ValueError, "Invalid min " + str(min)
            else:
                return [int(min)]


        
    def should_run(self, test_time):
        """Determine if the cron should run at test_time.
        test_time can be a unix timestamp (an int) or datetime object
        Note that seconds are ignroed
        """
        
        if not isinstance(test_time, datetime.datetime):
            test_time = datetime.datetime.fromtimestamp(test_time)
        test_time = test_time.replace(second=0, microsecond=0)
        
        #check validity:
        if test_time.month not in self.cron_compiled['mon']:
            return False
        if test_time.day not in self.cron_compiled['day']:
            return False
        if test_time.hour not in self.cron_compiled['hour']:
            return False
        if test_time.minute not in self.cron_compiled['min']:
            return False
        #dow is special case as it may be stored as a '*' for everything
        dow = self.cron_compiled['dow']
        if str(dow) == '*':
            return True
        #use isoweekday to match bsd cron standard where sunday is 0
        return test_time.isoweekday() in self.cron_compiled['dow']


"""unit test code"""
if __name__ == '__main__':
    k = CronTime('* * * * *') 
    assert(k.should_run(843848))
    assert(k.should_run(datetime.datetime(year=2010,month=2,day=3,hour=10,minute=0)))
    k = CronTime('1/3 * * * *') 
    assert(k.should_run(datetime.datetime(year=2010,month=2,day=3,hour=10,minute=0)) == False)
    assert(k.should_run(datetime.datetime(year=2010,month=2,day=3,hour=10,minute=31)) == True)
    k = CronTime('1-34/5 */2 1-10 mar *') 
    assert(k.should_run(datetime.datetime(year=2010,month=3,day=3,hour=10,minute=0)) == False)
    assert(k.should_run(datetime.datetime(year=2010,month=3,day=3,hour=10,minute=30)) == False)
    assert(k.should_run(datetime.datetime(year=2010,month=3,day=3,hour=10,minute=36)) == False)
    assert(k.should_run(datetime.datetime(year=2010,month=3,day=3,hour=11,minute=31)) == False)
    assert(k.should_run(datetime.datetime(year=2010,month=3,day=3,hour=10,minute=31)) == True)
    assert(k.should_run(datetime.datetime(year=2012,month=3,day=3,hour=10,minute=31)) == True)
    assert(k.should_run(datetime.datetime(year=2012,month=4,day=3,hour=10,minute=31)) == False)
    
    #day of week
    k = CronTime('1-34/5 */2 * 0/4 wed') 
    assert(k.should_run(datetime.datetime(year=2010,month=8,day=3,hour=10,minute=31)) == False)
    assert(k.should_run(datetime.datetime(year=2010,month=8,day=4,hour=10,minute=31)) == True)
    assert(k.should_run(datetime.datetime(year=2010,month=8,day=5,hour=10,minute=31)) == False)
    
    k = CronTime('1-34/5 */2 * 0/4 3-4') 
    assert(k.should_run(datetime.datetime(year=2010,month=8,day=3,hour=10,minute=31)) == False)
    assert(k.should_run(datetime.datetime(year=2010,month=8,day=4,hour=10,minute=31)) == True)
    assert(k.should_run(datetime.datetime(year=2010,month=8,day=5,hour=10,minute=31)) == True)
    assert(k.should_run(datetime.datetime(year=2010,month=8,day=6,hour=10,minute=31)) == False)



"""
Queues provide an interface for flow-based programming on PiCloud.

Please see our `Queue documentation <http://docs.scivm.com/queue.html>`_.
"""

from __future__ import with_statement
from __future__ import absolute_import
"""
Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2013 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see
http://www.gnu.org/licenses/lgpl-2.1.html
"""

"""Dev beware: list is defined in this module; list will not map to builtin list. Use builtin_list instead"""

import sys
import copy
import json
import time
import uuid
import base64
import string
import types
import numbers

from itertools import islice
from functools import partial
import traceback
import threading
from __builtin__ import list as builtin_list

from functools import wraps
from .serialization import pickledebug as saving_pickle
import cPickle as loading_pickle
from inspect import getargspec

__httpConnection = None
__url = None

from .util import  min_args, max_args

from .util.zip_packer import Packer
from .cloud import CloudException
from . import _getcloudnetconnection, _getcloud


import logging
cloudLog = logging.getLogger('Cloud.queue')

_queue_create_query = 'queue/'
_queue_list_query = 'queue/'
_queue_count_query = 'queue/{name}/count/'
_queue_exists_query = 'queue/{name}/exists/'
_queue_info_query = 'queue/{name}/'
_queue_delete_query = 'queue/{name}/delete/'
_queue_messages_push_query = 'queue/{name}/push/'
_queue_messages_pop_query = 'queue/{name}/pop/'
_queue_messages_pop_tickets_query = 'queue/{name}/pop_tickets/'
_queue_messages_ack_query = 'queue/{name}/ack/'
_queue_messages_update_deadline_query = 'queue/{name}/update_deadline/'
_queue_attach_query = 'queue/{name}/attach/'
_queue_detach_query = 'queue/{name}/detach/'


def _strip_unicode(s):
    """remove any unicode strings from s"""
    return s.decode('ascii', 'replace').encode('ascii', 'replace')

def _post(conn, url, post_values, headers={}):
    """Use HttpConnection *conn* to issue a post request to *url*
    with values *post_values*. HTTP *headers* may be added as well."""

    url = _strip_unicode(url)

    if 'success_action_redirect' in headers:
        headers['success_action_redirect'] = _strip_unicode(headers['success_action_redirect'])
    if post_values and 'success_action_redirect' in post_values:
        post_values['success_action_redirect'] = _strip_unicode(post_values['success_action_redirect'])

    cloudLog.debug('post url %s with post_values=%s. headers=%s'
                   % (url, post_values, headers))
    response =  conn.post(url, post_values, headers, use_gzip=False)

    return response


def list():
    """Returns a list of CloudQueue objects representing all registered
    Queues.

    Registration is lazy. See get() for more info."""

    conn = _getcloudnetconnection()

    resp = conn.send_request(_queue_list_query, None)

    return [CloudQueue(q) for q in resp['queues']]


def get(name):
    """Returns a CloudQueue object with *name*.

    Note that returned CloudQueue is lazily registered on PiCloud. Until an
    operation with the queue is executed, ``exists(name)`` will return False,
    and the queue will not appear in ``list()``"""

    return CloudQueue(name)

def exists(name):
    """Returns True if a CloudQueue of *name* is registered."""

    conn = _getcloudnetconnection()

    resp = conn.send_request(_queue_exists_query.format(name=name), None)

    return resp['exists']


class CloudQueue(object):
    """A CloudQueue is a queue registered on PiCloud.

    All messages in the queue are stored on PiCloud. Any machine can pop
    these messages, or push more, by creating a CloudQueue with the same
    ``name``."""

    # if queue should raise messages on message exceptions
    _raise_message_exceptions = True

    # valid letters for the name of a queue
    _valid_chars_in_name = string.letters + string.digits + '-_'

    # maximum number of bytes a message can be (larger messages
    # will be stored in bucket)
    _max_encoded_message_length = 64000

    # maximum number of messages to push in a single push
    _queue_push_limit = 800

    # maximum number of seconds to wait for an ack before msg is re-added to queue
    _max_deadline = 43200

    # minimum number of seconds
    _min_deadline = 3


    def __init__(self, name):
        """Creates a new CloudQueue. Use get() instead."""

        for c in name:
            if c not in self._valid_chars_in_name:
                raise ValueError('name must only container alphanumeric, '
                                 'underscore, or hyphen characters')

        if len(name) >= 40:
            raise ValueError('Length of name (%s) exceeded max of 40 characters'
                              % len(name))

        # name of the queue
        self.name = name

    def count(self):
        """Returns an approximation of the number of messages in the queue."""

        conn = _getcloudnetconnection()

        resp = conn.send_request(_queue_count_query.format(name=self.name), None)

        return resp['count']

    def info(self):
        """Returns a dictionary describing the queue (notably attachment information)"""

        conn = _getcloudnetconnection()

        resp = conn.send_request(_queue_info_query.format(name=self.name), None)

        return resp['info']

    def delete(self):
        """Delete the queue. All messages will be lost."""

        conn = _getcloudnetconnection()

        resp = conn.send_request(_queue_delete_query.format(name=self.name), None)

        return True

    def detach(self):
        """Detaches any attached message handler"""

        conn = _getcloudnetconnection()

        resp = conn.send_request(_queue_detach_query.format(name=self.name), None)

        return True

    def _encode_message(self, message, cur_retry=0, max_retries=None):
        """Encode a message in our own structure to include metadata.

        Our structure uses a top-level dictionary that includes keys for
        the message, which has been serialized, and datatype,which describes
        what serialization method was used."""

        json_d = None

        base_dictionary = {'cur_retry' : cur_retry,
                           'max_retries' : max_retries}

        try:

            # by default, try to json-encode the message
            # NOTE: do we really want dict keys to be converted to strings?
            # this means the message that was pushed won't be identical to the
            # message that is popped.
            serialized_message = json.dumps(message)
            datatype = 'json'
            json_d = {'message' : serialized_message,
                      'datatype' : datatype}

        except (TypeError, UnicodeDecodeError):
            # errors json.dumps may raise

            try:
                # fallback if JSON fails -- use pickle
                serialized_message = saving_pickle.dumps(message, 2)
                datatype = 'base64_python_pickle'

                # only try to encode it if it is small enough
                if len(serialized_message) < self._max_encoded_message_length:
                    json_d = {'datatype': datatype,
                              'message': base64.b64encode(serialized_message)}

            except:
                cloudLog.exception('Could not pickle message')
                raise

        if json_d:
            json_d.update(base_dictionary)
            encoded_message = json.dumps(json_d)

        if not json_d or len(encoded_message) > self._max_encoded_message_length:
            from . import bucket

            # if the message is too big, store it in user's bucket
            obj_key = 'queue/{name}/{uuid}'.format(name=self.name,
                                                   uuid=uuid.uuid4())
            bucket.putf(serialized_message, obj_key)

            if datatype == 'base64_python_pickle':
                datatype = 'python_pickle'

            json_d = {'datatype': datatype,
                      'redirect': 'bucket',
                      'key': obj_key}
            json_d.update(base_dictionary)
            encoded_message = json.dumps(json_d)

        return encoded_message

    @classmethod
    def _split_query(cls, query_func, messages):
        """Split messages into chunks of size ``_queue_push_limit``. For
        each chunk, apply *query_func*."""

        try:
            messages_iter = iter(messages)
        except TypeError:
            raise TypeError('messages must be an iterable')

        while True:
            batched_messages = builtin_list(islice(messages_iter, cls._queue_push_limit))
            if not batched_messages:
                break

            query_func(batched_messages)

        return True

    def push(self, messages, delay=0):
        """Put *messages*, a list of objects, into queue. If *delay*
        is specified, the function will return immediately, but the messages
        will not be available to readers (poppers) for *delay* seconds.

        Each object in *messages* must be pickle-able.

        Note that if the picked message exceeds 64kb, it will be temporarily saved
        to your cloud.bucket under the queue/ prefix."""

        return self._push(messages,delay)

    def _push(self, messages, delay=0, cur_retry=0, max_retries=None):
        """internal push to support propogating retry information"""

        if not isinstance(delay, numbers.Integral):
            raise TypeError('delay must be an integer')

        if not 900 >= delay >= 0:
            raise ValueError('delay must be between 0 and 900 seconds inclusive.')

        if not hasattr(messages, '__iter__'):
            raise ValueError('messages must be an iterable')

        messages_generator = (self._encode_message(message, cur_retry, max_retries)
                              for message in messages)
        query_func = partial(self._low_level_push,delay=delay)

        return self._split_query(query_func, messages_generator)

    def _low_level_push(self, encoded_messages, delay=0):
        """Performs the HTTP request to push encoded messages."""
        conn = _getcloudnetconnection()
        resp = conn.send_request(_queue_messages_push_query.format(name=self.name),
                         {'message': encoded_messages, 'delay': delay})
        return True

    def _decode_message(self, encoded_message):
        """Decodes the message to make it Python-friendly.
        Returns tuple of decoded message and redirect-key (if any) for cleanup purposes
        """

        redirect_key = None

        # first decode top-level structure
        if isinstance(encoded_message, basestring):
            message_data = json.loads(encoded_message)
        elif isinstance(encoded_message, dict):
            message_data = encoded_message
        else:
            raise TypeError('encoded_message must be a json string or dictionary')

        # next, decode message payload
        if isinstance(message_data, dict) and message_data.get('datatype'):

            redirect = message_data.get('redirect')
            datatype = message_data.get('datatype')

            if redirect and redirect == 'bucket':
                from . import bucket
                redirect_key = message_data['key']
                message_payload = bucket.getf(redirect_key).read()
            elif redirect:
                # only support redirecting to buckets right now
                raise RuntimeError('Do not understand redirect %s' % redirect)
            else:
                message_payload = message_data['message']

            if datatype == 'python_pickle':
                return loading_pickle.loads(message_payload), redirect_key
            elif datatype == 'base64_python_pickle':
                return loading_pickle.loads(base64.b64decode(message_payload)), redirect_key
            elif datatype == 'json':
                return json.loads(message_payload), redirect_key
            else:
                raise Exception('Unknown datatype %s' % datatype)

        else:
            return message_data, redirect_key

    def _clean_message(self, redirect_key):
        """perform necessary clean up after processing a message
        Supports multiple redirect_keys
        """
        if redirect_key:
            from . import bucket
            bucket.remove(redirect_key)

    def _low_level_pop(self, max_count, timeout):
        """Pops a list of json-encoded messages"""

        conn = _getcloudnetconnection()

        resp = conn.send_request(_queue_messages_pop_query.format(name=self.name),
                                 {'max_count': max_count, 'timeout': timeout})

        return resp['messages']

    def _low_level_pop_tickets(self, max_count, timeout, deadline):
        """pops a list of json encoded messages"""

        conn = _getcloudnetconnection()

        resp = conn.send_request(_queue_messages_pop_tickets_query.format(name=self.name),
                                 {'max_count': max_count,
                                  'timeout': timeout,
                                  'deadline': deadline})

        return resp['messages']


    def _pop_validator(self, max_count, timeout):
        """Checks to ensure arguments to pop are valid."""

        if not isinstance(max_count, numbers.Integral):
            raise TypeError('max_count must be an integer')

        if not (10 >= max_count > 0):
            raise ValueError('max_count must be between 1 and 10 inclusive')

        if not isinstance(timeout, numbers.Integral):
            raise TypeError('timeout must be an integer')

        if not (20 >= timeout >= 0):
            raise ValueError('timeout must be between 0 and 20 inclusive')


    def pop(self, max_count=10, timeout=20):
        """Pops at most *max_count* messages from the queue. Returns a
        list of popped messages. Note that the non-FIFO order of the pop
        is intentional.

        Blocks until at least one message is available, or until *timeout*
        seconds elapses, at which point an empty list is returned.

        *timeout* must be specified as an integer between 0 and 20 inclusive.
        *max_count* must be specified as an integer between 1 and 10 inclusive."""

        self._pop_validator(max_count, timeout)

        results_data = self._low_level_pop(max_count, timeout)

        messages = []

        for datum in results_data:
            try:
                decoded_msg, redirect_key = self._decode_message(datum)
                messages.append(decoded_msg)
            except:
                cloudLog.exception('Could not decode a message')
                if not self._raise_message_exceptions:
                    sys.stderr.write('Could not decode a message.\n')
                    traceback.print_exc()
                    sys.stderr.write('\n')
                else:
                    raise
            else:
                self._clean_message(redirect_key)

        return messages

    def pop_tickets(self, max_count=10, timeout=20, deadline=30):
        """Pops at most *max_count* messages from the queue. Returns a list
        of MessageTickets.

        Semantics are similar to pop, except that instances of MessageTicket
        are returned which encapsulates the actual message.

        All MessageTickets must be acknowledged within *deadline* seconds.
        If they are not the message will be re-added to the queue.

        Using pop_tickets and MessageTicket.ack() (or ack_messages) allows
        you to build highly robust messaging systems where a message will
        go unprocessed."""

        self._pop_validator(max_count, timeout)

        if not isinstance(deadline, numbers.Integral):
            raise TypeError('deadline must be an integer')

        if not (self._max_deadline >= deadline >= self._min_deadline):
            raise ValueError('deadline must be between %s and %s inclusive'
                             % (self._min_deadline, self._max_deadline))


        results_data = self._low_level_pop_tickets(max_count, timeout, deadline)

        messages = []

        for decoded_wrapper_dct in results_data:
            try:
                decoded_message, redirect_key = self._decode_message(decoded_wrapper_dct)
                ticket = MessageTicket(self, decoded_message, decoded_wrapper_dct['ticket'],
                                       deadline, redirect_key=redirect_key)
                messages.append(ticket)
            except:
                cloudLog.exception('Could not decode a message')
                if not self._raise_message_exceptions:
                    sys.stderr.write('Could not decode a message.\n')
                    traceback.print_exc()
                    sys.stderr.write('\n')
                else:
                    raise

        return messages


    def attach(self, message_handler, output_queues=[], iter_output=False,
                readers_per_job=1, max_parallel_jobs=1, retry_on=[],
                max_retries=None, retry_delay=None, on_error={}, **kwargs):
        """Register *message_handler* as the processor for this queue on PiCloud.

        If ``message_handler`` is a function (or any callable), then it will be
        invoked for each message, with the message as the only argument.

        The queue processing will occur in the context of a job. The job automatically
        ends if no data is available in the queue for 20 seconds.

        Alternatively, ``message_handler`` may be an instance if it is desired to maintain state
        between message processing. In this case, the mssage_handler function will be the
        instance's method "message_handler()" which takes a single argument (a message).
        The instance may optionally provide ``pre_handling`` and ``post_handling`` methods to manage
        startup and cleanup respectively. ``pre_handling`` is called once before the first
        message is processed by the wrapping job, and ``post_handling`` is called once after the last
        message is processed by the wrapping job.

        By default, the return value of the message_handler is pushed to each queue in ``output_queues``.
        If the return value is None, no message is pushed.

        If ``iter_output`` is set to True, the return value of the message_handler will be considered
        an iterable.  Each element of the iterator (including any None values) will be pushed to
        each queue in ``output_queues``.

        ``readers_per_job controls`` the number of messages can be processed by a job in parallel.

        ``max_parallel_jobs`` is the maximum number of jobs that can run in parallel to process your queue.
            Set to 0 to disable processing

        ``retry_on``, ``max_retries``, ``retry_delay`` and ``on_error`` all relate to error handling.
        Please see our online documentation: http://docs.scivm.com/queue.html#error-handling

        Certain special *kwargs* associated with cloud.call can be used for your *message_handler*:

        * _cores:
            Set number of cores your job will utilize. See http://docs.scivm.com/primer.html#choose-a-core-type/
            In addition to having access to more CPU cores, the amount of RAM available will grow linearly.
            Possible values for ``_cores`` depend on what ``_type`` you choose:

            * c1: 1
            * c2: 1, 2, 4, 8
            * f2: 1, 2, 4, 8, 16
            * m1: 1, 2, 4, 8
            * s1: 1
        * _env:
            A string specifying a custom environment you wish to run your job within.
            See environments overview at http://docs.scivm.com/environment.html
        * _label:
            A user-defined string label that is attached to the jobs. Labels can be
            used to filter when viewing jobs interactively (i.e. on the PiCloud website).
        * _os_env_vars:
            List of operating system environment variables that should be copied to PiCloud from your system
            Alternatively a dictionary mapping the environment variables to the desired values.
        * _priority:
            A positive integer denoting the jobs' priority. PiCloud tries to run jobs
            with lower priority numbers before jobs with higher priority numbers.
        * _profile:
            Set this to True to enable profiling of your code. Profiling information is
            valuable for debugging, and finding bottlenecks.
            **Warning**: Profiling can slow your job significantly; 10x slowdowns are known to occur
        * _restartable:
            In the rare event of hardware failure while a job is processing, this flag
            indicates whether the job can be restarted. By default, this is True. Jobs
            that modify external state (e.g. adding rows to a database) should set this
            False to avoid potentially duplicate modifications to external state.
        * _type:
            Choose the type of core to use, specified as a string:

            * c1: 1 compute unit, 300 MB ram, low I/O (default)
            * c2: 2.5 compute units, 800 MB ram, medium I/O
            * f2: 5.5 compute units, 3.75 GB ram, high I/O, hyperthreaded core
            * m1: 3.25 compute units, 8 GB ram, high I/O
            * s1: Up to 2 compute units (variable), 300 MB ram, low I/O, 1 IP per core

            See http://www.scivm.com/pricing/ for pricing information
        * _vol:
            A string or list of strings specifying a volume(s) you wish your jobs to have access to.

        """

        # TODO: Bring back batch_size to the public API. If we do, it will
        # represent the number of messages an attachment should get at once.

        cloud = _getcloud()
        params = cloud._getJobParameters(message_handler, kwargs,
                                         ignore=['_depends_on', '_depends_on_errors',
                                                 '_max_runtime', '_fast_serialization',
                                                 '_kill_process'])

        return _low_level_attach(self, message_handler, iter_output, output_queues, None,
                                 readers_per_job, max_parallel_jobs, retry_on,
                                 max_retries, retry_delay, on_error, params)


    def __repr__(self):
        return "CloudQueue('{name}')".format(name=self.name)

    def __eq__(self, other):
        if not isinstance(other, CloudQueue):
            return False
        return self.name == other.name

    def _validate_tickets(self, message_tickets):
        if not hasattr(message_tickets, '__iter__'):
            raise TypeError('message_tickets must be an iterable')

        for ticket in message_tickets:
            if isinstance(ticket, MessageTicket):
                if ticket.queue and ticket.queue != self:
                    raise ValueError('MessageTicket is not associated with this queue')
            elif not isinstance(ticket, basestring):
                raise TypeError('message_tickets elements must be MessageTicket or strings')


    def ack(self, message_tickets):
        """Ack(nowledge) the ticket - and successful handling of - multiple
        *message_tickets*. If acknowledge is not called, the message will be
        automatically placed back into the queue when deadline is reached.

        *message_tickets* can be a MessageTicket or the actual ticket id.
        """

        self._validate_tickets(message_tickets)

        conn = _getcloudnetconnection()

        def query_func(batched_tickets):
            conn.send_request(_queue_messages_ack_query.format(name=self.name),
                              {'ticket' : batched_tickets})

        ticket_generator = ((mt.ticket if isinstance(mt, MessageTicket) else mt)
                            for mt in message_tickets)

        retval = self._split_query(query_func, ticket_generator)
        if retval:
            # clean up
            redirect_generator = ((mt.redirect_key for mt in message_tickets
                                   if isinstance(mt,MessageTicket) and mt.redirect_key))
            for redirect_key in redirect_generator:
                self._clean_message(redirect_key)

    def update_deadline(self, message_tickets, new_deadline):
        """Modify the deadline of multiple not-yet ack-ed *message_tickets*
        After invoking this function, the message_tickets' deadline will
        occur in *new_deadline* seconds relative to NOW.

        *message_tickets* can be a message_ticket or the ticket id.
        """

        self._validate_tickets(message_tickets)
        if not isinstance(new_deadline, numbers.Integral):
            raise TypeError('new_deadline must be an integer')
        if not (self._max_deadline >= new_deadline >= self._min_deadline):
            raise ValueError('new_deadline must be between 3 and 43200 inclusive')

        conn = _getcloudnetconnection()
        def query_func(batched_tickets):
            conn.send_request(_queue_messages_update_deadline_query.format(name=self.name),
                              {'ticket' : batched_tickets, 'deadline' : new_deadline})

        ticket_generator = ((mt.ticket if isinstance(mt, MessageTicket) else mt)
                            for mt in message_tickets)
        now = time.time()
        self._split_query(query_func, ticket_generator)

        for message in message_tickets:
            if isinstance(message, MessageTicket):
                message._set_deadline(new_deadline, now)

class MessageTicket(object):
    """A MessageTicket includes information and functions for
    acknowleding a message, or changing its deadline.
    It also includes the message itself.

    To access the message, read the ``message`` attribute.
    To access ticket id, read the ``ticket_id`` attribute
    """

    # queue that message was popped from
    queue = None

    # actual message
    message = ''

    # ticket uniquely represents the delivery of the message
    ticket = ''

    # number of seconds from receipt of message until this message is
    # automatically re-added to the queue
    deadline = None

    # time.time() when object was created (needed for deadline)
    _creation_time = None

    # optional redirect_key (external location to be cleaned up when message is acknowledged)
    redirect_key = None


    def __init__(self, queue, message, ticket, deadline, creation_time=None, redirect_key=None):
        """Creates a new MessageTicket. Do not use directly, use
        ``pop_tickets`` instead."""

        self.queue = queue
        self.message = message
        self.ticket = ticket
        self._set_deadline(deadline, creation_time)
        self.redirect_key = redirect_key

    def _set_deadline(self, deadline, creation_time):
        """Internal function for modifying the deadline"""

        self.deadline = deadline
        self._creation_time = creation_time or time.time()

    def ack(self):
        """Acknowledge this ticket. For batch acknowledgements,
        see CloudQueue.acknowledge_message"""

        return self.queue.ack([self])

    def time_until_deadline(self):
        """Returns how much time (in seconds) is left until the message's
        deadline."""
        return self._creation_time + self.deadline - time.time()

    def update_deadline(self, new_deadline):
        """If the deadline you initially set it insufficient (the message cannot
        be reliably purged from the queue), update its deadline. The *new_deadline*
        should be the number of seconds in the future (relative to now) before
        the message is re-added to the queue automatically."""

        return self.queue.update_deadline([self], new_deadline)

    def __repr__(self):
        return "MessageTicket('{message}')".format(message=self.message)

class Retry(Exception):
    """Raise this Exception within an attachment to abort and retry the message

    Within the constructor, the delay muyst be set to specify how many seconds
        to wait until the retried message can be dequeued.
        max_retries must also be set to specify how many times the message can be retried
    """

    # Parameters for retried message itself
    delay = None  # If set, override default delay
    max_retries = None # if set, override default max_retries


    # information
    inner_excp_value = None # inner exception that triggered this retry
    traceback_info = None # traceback of the inner exception
    message = '' # some optional extra message


    def __init__(self, delay, max_retries, message='',
                 inner_excp_value=None, traceback_info=None, **kwargs):

        if not isinstance(delay, (numbers.Integral,types.NoneType)):
            raise TypeError('delay must be an integer')

        if not isinstance(max_retries, (numbers.Integral,types.NoneType)):
            raise TypeError('max_retries must be an integer')

        auto_excp_type, auto_excp_value, auto_traceback = sys.exc_info()
        if not traceback_info:
            traceback_info = ''.join(traceback.format_tb(auto_traceback))
        if not inner_excp_value:
            inner_excp_value = auto_excp_value

        self.traceback_info = traceback_info
        self.inner_excp_value = inner_excp_value
        self.message = message
        self.delay = delay
        self.max_retries = max_retries

        super(Retry, self).__init__(delay, max_retries, message,
                                    inner_excp_value, traceback_info, **kwargs)

    def __str__(self):
        if self.inner_excp_value:
            return 'Retry(%s) on %s: %s:\n%s' % (self.message, type(self.inner_excp_value),
                                                                    self.inner_excp_value,
                                                                    self.traceback_info)
        else:
            return 'Retry(%s)' % self.message

    def __repr__(self):
        return self.__str__()

class MaxRetryException(Retry):
    """Signals maximum retries exceeded"""
    pass


def _validate_action_dct(action_dct):
    """Ensures that an action_dct (on_error) dict is well-formed"""

    for e, action in action_dct.items():
        if not issubclass(e, Exception):
            raise ValueError('Key is not an Exception')
        if 'queue' not in action:
            raise ValueError('CloudQueue not specified for action')
        elif not isinstance(action['queue'], CloudQueue):
            raise ValueError('Specified queue must be CloudQueue object')
        if 'delay' in action and not 900 >= action['delay'] >= 0:
            raise ValueError('delay must be between 0 and 900 seconds inclusive.')


def _stringify_action_dct(action_dct):
    """Creates a version of *action_dct* that is composed purely of
    strings. This makes it JSON-able."""

    action_dct = copy.deepcopy(action_dct)
    for e, action in action_dct.items():
        action_dct[e.__name__] = action
        del action_dct[e]
        action['queue'] = action['queue'].name

    return action_dct

def _low_level_attach(queue, message_handler, expand_iterable_output, output_queues,
                      batch_size, readers_per_job, max_parallel_jobs, retry_on,
                      max_retries, retry_delay, on_error, params):
    """Does the grunt work in attaching the *message_handler* to the queue."""

    if not isinstance(readers_per_job, numbers.Integral):
        raise TypeError('readers_per_job must be an integer')
    if not isinstance(max_parallel_jobs, numbers.Integral):
        raise TypeError('timeout must be an integer')

    if not (max_parallel_jobs >= 0):
        raise ValueError('max_parallel_jobs must be greater than or equal to 0')

    if not (128 >= readers_per_job >= 1):
        raise ValueError('readers_per_job must be between 1 and 128 inclusive')

    if not hasattr(retry_on, '__iter__'):
        retry_on = [retry_on]

    if retry_on:
        if max_retries is None:
            raise ValueError('max_retries must be set if retry_on is specified')
        if retry_delay is None:
            raise ValueError('retry_delay must be set if retry_on is specified')

        if not isinstance(max_retries, numbers.Integral):
            raise TypeError('max_retries must be an integer')
        if not isinstance(retry_delay, numbers.Integral):
            raise TypeError('retry_delay must be an integer')


    _validate_action_dct(on_error)

    if output_queues:
        if not hasattr(output_queues, '__iter__'):
            output_queues = [output_queues]

        if len(output_queues) > 50:
            raise ValueError('Cannot specify more than 50 output queues.')

        for output_queue in output_queues:
            if not isinstance(output_queue, CloudQueue):
                raise ValueError('output queues must be CloudQueue instances')

    implements_interface = False

    if not callable(message_handler):
        # if obj has a message_handler member, it's all good
        # if obj has a message_handler member, but not callable, error
        # if obj doesn't have a message_handler member, error with callable needed

        if not hasattr(message_handler, 'message_handler'):
            raise TypeError('cloud.queue.attach *message_handler* '
                            'argument (%s) is not callable'
                            % (str(message_handler)))

        elif not callable(getattr(message_handler, 'message_handler')):
            raise TypeError('cloud.queue.attach *message_handler* is class without (%s) callable'
                            'message_handler() member function'  % (str(message_handler)))

        implements_interface = True

    # TODO: batch_size can be determined automatically by measuring the average
    # time and variance it takes to process a message. Assume that a batch pop
    # takes ~100ms, and then do the match.
    # In theory, with more parallel jobs, lower batch size (heroku problem)
    if not batch_size:
        batch_size = 10  # maximum possible

    try:
        if implements_interface:
            min_arg = min_args(message_handler.message_handler)
            max_arg = max_args(message_handler.message_handler)
        else:
            min_arg = min_args(message_handler)
            max_arg = max_args(message_handler)
    except TypeError:
        pass #type can't be sanity checked.. let it through
    else:
        # TODO: Validate special arguments it accepts
        if max_arg < 1:
            raise ValueError('message_handler function must accept 1 (required) argument. %s only accepts %s' \
                             % (str(message_handler), max_arg))

    attachment = _create_attachment(queue, message_handler, expand_iterable_output, output_queues,
                                    batch_size, readers_per_job, retry_on, max_retries,
                                    retry_delay, on_error)

    cloud = _getcloud()
    conn = _getcloudnetconnection()

    os_env_vars = params.pop('os_env_vars', None)
    smessage_handler, sarg, logprefix, logcnt = cloud.adapter.cloud_serialize(attachment, 2, [],
                                                                   logprefix='queue.',
                                                                   os_env_vars=os_env_vars)

    conn._update_params(params)

    cloud.adapter.dep_snapshot()

    data = Packer()
    data.add(smessage_handler)
    params['data'] = data.finish()
    params['output_queue'] = [output_queue.name for output_queue in output_queues]
    params['batch_size'] = batch_size
    params['readers_per_job'] = readers_per_job
    params['max_parallel_jobs'] = max_parallel_jobs
    params['max_retries'] = max_retries
    params['retry_delay'] = retry_delay
    params['on_error'] = json.dumps(_stringify_action_dct(on_error))
    params['retry_on'] = json.dumps([excp.__name__ for excp in retry_on])

    conn.send_request(_queue_attach_query.format(name=queue.name), params)

def _create_attachment(input_queue, message_handler, expand_iterable_output, output_queues,
                       batch_size, readers_per_job, retry_on=[], max_retries=None, retry_delay=None,
                       on_error={}, _job_shutdown_timeout=20):
    """Creates the function to be executed on PiCloud."""

    def attachment():
        from scicloud.transport import queue_attachment

        return queue_attachment.launch_attachment(input_queue, message_handler, expand_iterable_output,
                                                  output_queues, batch_size, readers_per_job,
                                                  retry_on, max_retries, retry_delay, on_error,
                                                  _job_shutdown_timeout)

    if callable(message_handler):
        return wraps(message_handler)(attachment)
    else:
        return wraps(message_handler.message_handler)(attachment)
